# fmt: off

from textwrap import dedent
from time import time_ns
from typing import TYPE_CHECKING, Self
from collections.abc import Iterable

from ..core.sentinels import NO_VALUE, NoValue

from .repl_alias import ReplException
from .repl_eval_info import ReplEvaluationInfo
from .repl_hooks import SystemHooks
from .repl_error import ReplError

__all__ = [
    'ReplEvaluationData',
    'NO_VALUE'
]


################################################################################

if TYPE_CHECKING:
    from .repl import BaseRepl
    from .repl_code import ReplCode
    from .repl_vars import VarsDelta


################################################################################

class ReplEvaluationData:
    """
    Represents the results of evaluating a single code block within a `Repl`.

    This is created by `Repl.run_code()`, and populated as execution proceeds.

    Fields:
    `code`:        `ReplCode` object containing the raw and compiled source
    `output`:       string containing all output emitted during execution
    `error:         info about a terminating exception, if any
    `return_value`: value explicitly returned by a top-level `return foo`, if any
    `out_value`:    if last statement is expression, its value
    `out_str`:      formatted version of the above
    `start_pos`:    position in the repl output stream when evaluation started
    `stop_pos`:     ... when evaluation stopped
    `start_time`:   `time_ns()` when evaluation started
    `stop_time`:    ... when evaluation stopped
    `imported`:     list of module names imported during evaluation
    `vars_delta`:   how the set of local vars changed during execution

    `returned` and `last` are set to `NO_VALUE` initially.

    The following properties are computed from the above:
    `has_error`
    `has_uncaught_error`
    `has_raised_error`
    `has_return_value`
    `has_out_value`
    `duration`
    `traceback_str`

    A summary of the execution in the form of a `ReplEvaluationInfo` can
    be obtained via `to_summary`.
    """

    code:        'ReplCode'
    output:       str | None
    error:        ReplError | None

    return_value: object | NoValue
    out_value:    object | NoValue
    out_str:      str | None

    start_pos:    int
    stop_pos:     int
    start_time:   int
    stop_time:    int

    imported:     list[str] | None
    locals_delta: 'VarsDelta'

    def __init__(self, code: 'ReplCode') -> None:
        self.code = code
        self.output = None
        self.error = None
        self.return_value = NO_VALUE
        self.out_value = NO_VALUE
        self.out_str = None
        self.start_time = 0
        self.stop_time = 0
        self.output_pos = 0
        self.imported = []

    # --------------------------------------------------------------------------

    def __debug_info_str__(self) -> str:
        return f'{self.code.source_hash:10x}'

    @property
    def is_async(self) -> bool:
        return self.code.is_async

    @property
    def has_finished(self) -> bool:
        return self.stop_time != 0

    @property
    def has_error(self) -> bool:
        return self.error is not None

    @property
    def has_uncaught_error(self) -> bool:
        return self.error is not None and self.error.uncaught

    @property
    def traceback_str(self) -> str | None:
        return self.error.traceback if self.error is not None else None

    @property
    def has_raised_error(self) -> bool:
        return self.error is not None and self.error.raised

    @property
    def has_return_value(self) -> bool:
        return self.return_value is not NO_VALUE

    @property
    def has_out_value(self) -> bool:
        return self.out_value is not NO_VALUE

    @property
    def duration(self) -> float:
        return (self.stop_time - self.start_time) / 1e9

    # --------------------------------------------------------------------------

    def __repr__(self) -> str:
        return '\n'.join(self.repr_lines())

    def repr_lines(self) -> Iterable[str]:
        yield 'ReplExecution('
        yield f"\tcode      = '''{self.code.source!r}''',"
        if self.output:
            yield f"\toutput    = '''{self.output}''',"
        yield f"\tduration  = {self.duration}"
        yield ")\n"

    def print(self) -> None:
        print(repr(self))

    def console_lines(self) -> Iterable[str]:
        input_str = dedent(self.code.source).strip()
        first, *rest = input_str.splitlines()
        yield f'>>> {first}'
        for line in rest:
            yield f'... {line}'
        output_str = self.output.strip()
        yield from output_str.splitlines()

    def console_str(self) -> str:
        return '\n'.join(self.console_lines())

    # --------------------------------------------------------------------------

    def eval_sync(self, repl: 'BaseRepl') -> Self:
        """Run assuming there is no event loop (or it is not running)."""
        if self.is_async:
            loop = repl.get_loop()
            if loop.is_running():
                raise ReplException("can't run async repl code in sync mode if loop is already running")
            task = loop.create_task(self.__eval_coro(repl), name='<repl code>')
            loop.run_until_complete(task)
        else:
            self.__eval_sync(repl)
        return self

    async def eval_async(self, repl: 'BaseRepl') -> Self:
        """Run assuming there is a running event loop."""
        if self.is_async:
            loop = repl.get_loop()
            if loop is None:
                raise ReplException("can't run async repl code if repl has no loop set")
            task = loop.create_task(self.__eval_coro(repl), name='<repl code>')
            await task
        else:
            self.__eval_sync(repl)
        return self

    # --------------------------------------------------------------------------

    def __eval_sync(self, repl: 'BaseRepl') -> None:
        state = self.__start(repl)
        try:
            self.code.sync_function()
            self.__stop(repl, state, None)
        except BaseException as exc:
            self.__stop(repl, state, exc)

    async def __eval_coro(self, repl: 'BaseRepl') -> None:
        state = self.__start(repl)
        try:
            await self.code.async_function()
            self.__stop(repl, state, None)
        except BaseException as exc:
            self.__stop(repl, state, exc)

    # --------------------------------------------------------------------------

    def __start(self, repl: 'BaseRepl') -> SystemHooks:
        state = SystemHooks.get()
        self.locals_delta = repl.vars.local_deltas.open()
        self.start_pos = repl.get_output_pos()
        repl.hooks.set()
        self.start_time = time_ns()
        repl.evaluation_started(self)
        return state

    def __stop(self, repl: 'BaseRepl', state: SystemHooks, exception: BaseException | None) -> None:
        state.set()
        repl.vars.local_deltas.close(self.locals_delta)
        self.stop_time = time_ns()
        self.stop_pos = repl.get_output_pos()

        if isinstance(exception, ReturnExit):
            if self.return_value is NO_VALUE:
                raise ReplException("ReturnExit caught but return_value not set")
            repl.evaluation_returned(self, self.return_value)

        elif isinstance(exception, BaseException):
            self.error = error = ReplError(exception)
            repl.evaluation_errored(self, error)

        self.output = repl.get_output_since(self.start_pos)
        repl.evaluation_finished(self)

    # --------------------------------------------------------------------------

    def to_info(self) -> ReplEvaluationInfo:
        if self.stop_time == 0:
            raise ReplException("evaluation has not completed")
        exception_name = self.error.name if self.has_error else None
        added, changed, removed = self.locals_delta.to_tuples()
        return ReplEvaluationInfo(
            output=truncate_str(self.output),
            out_str=truncate_str(self.out_str),
            traceback_str=truncate_str(self.traceback_str),
            exception_name=exception_name,
            has_return_value=self.has_return_value,
            has_raised_error=self.has_raised_error,
            has_uncaught_error=self.has_uncaught_error,
            duration=self.duration,
            added_locals=added,
            changed_locals=changed,
            removed_locals=removed,
            imported=tuple(self.imported),
            metadata={}
        )


################################################################################

MAX_LEN = 10_000
TAIL_LINES = 5
TAIL_MAX_CHARS = 500


def truncate_str(text: str | None) -> str | None:
    if not isinstance(text, str):
        return None
    if len(text) <= MAX_LEN:
        return text

    truncated = text[MAX_LEN:]
    n_chars = len(truncated)
    n_lines = truncated.count('\n') + (1 if truncated and not truncated.endswith('\n') else 0)

    # Grab last few lines as preview
    lines = text.rstrip('\n').split('\n')
    tail = lines[-TAIL_LINES:]
    tail_text = '\n'.join(tail)
    if len(tail_text) > TAIL_MAX_CHARS:
        tail_text = '...' + tail_text[-(TAIL_MAX_CHARS - 3):]

    return (
        f"{text[:MAX_LEN]}\n\n"
        f"[ ... {n_chars} chars ({n_lines} lines) truncated - maximum REPL output exceeded, inspect smaller slices ]\n\n"
        f"{tail_text}"
    )

################################################################################

class ReturnExit(SystemExit):
    ...



# fmt: off

from .repl_eval_info import ReplEvaluationInfo
from .repl_callable_info import ReplCallableInfo
from .repl_var_info import VarKind, ReplVarInfo
from .repl_session_info import ReplSessionInfo, ReplRole, VALID_REPL_ROLES

__all__ = [
    'VarKind',
    'ReplEvaluationInfo',
    'ReplCallableInfo',
    'ReplVarInfo',
    'ReplSessionInfo',
    'ReplRole',
    'VALID_REPL_ROLES',
]



# ruff: noqa
# fmt: off

from msgspec import UNSET, EncodeError, Raw, Struct, UnsetType, json, msgpack  # noqa: F401

from ..__ import *
from .__json import *
from .__msgpack import *

from .codec import EncoderP, DecoderP, CodecP

# if TYPE_CHECKING:
#     from .all import *

def seq_shape(seq: tuple['Msg', ...]) -> str:
    n = len(seq)
    return len_shape(n) if n > 4 else commas(m.shape for m in seq)

def len_shape(n: int) -> str:
    return f'..{n}'

commas = ', '.join



# ruff: noqa
# fmt: off

from ..__ import *

if TYPE_CHECKING:
    from ..msg.codec import EncoderP, DecoderP, CodecP
    from ..msg.rpc_request import RequestMsg
    from ..msg.rpc_request_resource import *
    from ..msg.rpc_request_repl import *



# ruff: noqa
# fmt: off

import pytest

from agentica_internal.testing.examples.annos import TreeAlias
from agentica_internal.testing.warpc.type_anno_tests import (
    ANNOS,
    CLASSES,
    verify_annotation_equality,
    verify_class_annotation_equality,
    verify_normalizing_unions
)


@pytest.mark.asyncio
@pytest.mark.parametrize('annotation', ANNOS)
async def test_warp_type_annotation(annotation):
    if annotation is TreeAlias:
        pytest.skip()
    await verify_annotation_equality(annotation)


@pytest.mark.asyncio
@pytest.mark.parametrize('cls', CLASSES)
async def test_warp_class_annotation(cls):
    await verify_class_annotation_equality(cls)


def test_warp_normalizing_unions():
    verify_normalizing_unions()



import asyncio
from types import ModuleType

from agentica_internal.warpc.forbidden import *
from agentica_internal.warpc.worlds.all import DebugWorld

F_NAME = 'testing.forbidden'


async def verify_forbidden_module():
    class f_class: ...

    def f_func(): ...

    f_mod = ModuleType(F_NAME)

    f_class.__module__ = F_NAME
    f_func.__module__ = F_NAME

    blacklist_modules(F_NAME)

    async with DebugWorld.connected_pair(logging=False) as B:
        f_cls_b = await B(f_class)
        assert f_cls_b.__name__ == '<forbidden class>'

        f_fun_b = await B(f_func)
        assert f_fun_b.__name__ == '<forbidden function>'

        f_mod_b = await B(f_mod)
        assert f_mod_b.__name__ == '<forbidden module>'


async def verify_forbidden_whitelist():
    class w_class: ...

    def w_func(): ...

    w_mod = ModuleType(F_NAME + '.whitelisted')

    w_class.__module__ = F_NAME
    w_func.__module__ = F_NAME

    blacklist_modules('forbidden_module')
    whitelist_objects(w_class, w_func, w_mod)

    async with DebugWorld.connected_pair(logging=False) as B:
        w_cls_b = await B(w_class)
        assert w_cls_b.__name__ == 'w_class'

        w_fun_b = await B(w_func)
        assert w_fun_b.__name__ == 'w_func'

        w_mod_b = await B(w_mod)
        assert w_mod_b.__name__ == 'testing.forbidden.whitelisted'


if __name__ == '__main__':
    asyncio.run(verify_forbidden_module())
    asyncio.run(verify_forbidden_whitelist())



# fmt: off

from .__ import *
from .base import *

__all__ = [
    'CoroutineData',
]


################################################################################

class CoroutineData(ResourceData):
    __slots__ = 'name', 'qname'

    FORBIDDEN_FORM = forbidden_object

    name: str
    qname: optstr

    @classmethod
    def describe_resource(cls, coro: CoroutineT) -> 'CoroutineData':
        data = CoroutineData()
        data.name = getattr(coro, NAME, '')
        data.qname = getattr(coro, QUALNAME, '')
        return data

    def create_resource(self, handle: ResourceHandle) -> 'coroutine':
        handle.kind = Kind.Coroutine
        handle.keys = []
        handle.open = False
        handle.name = self.name
        vobj = coroutine(self.name, self.qname)
        obj_set(vobj, VHDL, handle)
        return vobj


################################################################################

# this 'coroutine' mimics `types.coroutine` as closely as possible.

class coroutine(Coroutine):
    __slots__ = NAME, QUALNAME, VHDL

    if TYPE_CHECKING:
        __name__: str
        __qualname__: str

    def __init__(self, name: str = '', qualname: str = ''):
        self.__name__ = name = name or '<coroutine>'
        self.__qualname__ = qualname or name

    def cr_origin(self) -> tuple[tuple[str, int, str], ...] | None:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallMethod(self, 'cr_origin'))

    def cr_suspended(self) -> bool:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallMethod(self, 'cr_suspended'))

    def close(self) -> None:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallMethod(self, 'close'))

    def __await__(self) -> A.Generator:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallMethod(self, '__await__'))

    def send(self, arg: Any, /) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallMethod(self, 'send', (arg,)))

    def throw(self, typ: BaseException,
                    val: None = None,
                    tb: S.TracebackT | None = ..., /) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallMethod(self, 'throw', (type, val, None,)))


coroutine.__module__ = 'virtual'



# fmt: off

__all__ = [
    'DiamondC',
    'DiamondT',
    'DiamondB',
    'DiamondL',
    'DiamondR',
    'DiamondBL',
    'DiamondBR',
    'DiamondTL',
    'DiamondTR',
    'DIAMOND_CLASSES',
    'DIAMOND_INSTANCES',
]

type NEG = None
type POS = None
type ZER = None


class DiamondC:
    """DiamondC docsting"""

    x: ZER
    y: ZER


class DiamondB(DiamondC): y: NEG
class DiamondT(DiamondC): y: POS
class DiamondL(DiamondC): x: NEG
class DiamondR(DiamondC): x: POS

class DiamondBL(DiamondB, DiamondL): pass
class DiamondBR(DiamondB, DiamondR): pass
class DiamondTL(DiamondT, DiamondL): pass
class DiamondTR(DiamondT, DiamondR): pass

DIAMOND_CLASSES: list[type] = [
    DiamondC,
    DiamondT,
    DiamondB,
    DiamondL,
    DiamondR,
    DiamondBL,
    DiamondBR,
    DiamondTL,
    DiamondTR,
]

DIAMOND_INSTANCES: list[object] = [
    DiamondC(),
    DiamondT(),
    DiamondB(),
    DiamondL(),
    DiamondR(),
    DiamondBL(),
    DiamondBR(),
    DiamondTL(),
    DiamondTR(),
]



# fmt: off

from builtins import type as cls_t, object as obj_t
from types import ModuleType as mod_t

from ..attrs import VHDL, DICT

__all__ = [
    'cls_get',  'obj_get',  'mod_get',
    'cls_set',  'obj_set',  'mod_set',
    'cls_handle', 'mod_handle', 'obj_handle',
    'cls_dict', 'mod_dict', 'obj_dict',
]


################################################################################

cls_get = cls_t.__getattribute__  # type: ignore
obj_get = obj_t.__getattribute__  # type: ignore
mod_get = mod_t.__getattribute__  # type: ignore

cls_set = cls_t.__setattr__       # type: ignore
obj_set = obj_t.__setattr__       # type: ignore
mod_set = mod_t.__setattr__       # type: ignore

def cls_handle(cls: cls_t):
    return cls_get(cls, VHDL)

def mod_handle(mod: cls_t):
    return mod_get(mod, VHDL)

def obj_handle(obj: obj_t):
    return obj_get(obj, VHDL)

def cls_dict(cls: cls_t):
    return cls_get(cls, DICT)

def mod_dict(mod: mod_t):
    return mod_get(mod, DICT)

def obj_dict(obj: obj_t):
    return obj_get(obj, DICT)



# fmt: off

from .base import *
from .request_resource import *
from .request_future import *
from .request_repl import *

__all__ = [
    'Request',

    'ReplInit',
    'ReplRequest',
    'ReplRunCode',
    'ReplCallMethod',

    'FutureRequest',
    'CancelFuture',
    'CompleteFuture',

    'ResourceRequest',
    'ResourceNew',
    'ResourceCallRequest',
    'ResourceCallFunction',
    'ResourceCallMethod',
    'ResourceCallSystemMethod',
    'ResourceAttrRequest',
    'ResourceGetAttr',
    'ResourceHasAttr',
    'ResourceDelAttr',
    'ResourceSetAttr',

    'GENERIC_RESOURCE_ERROR'
]



# fmt: off

import asyncio
from collections import Counter, OrderedDict

from agentica_internal.warpc.worlds.debug_world import DebugWorld


class sub_object(object):
    def name(self):
        return 'sub_object'


class sub_list(list):
    def name(self):
        return 'sub_list'


class sub_dict(dict):
    def name(self):
        return 'sub_dict'


class sub_set(set):
    def name(self):
        return 'sub_set'


class sub_Counter(Counter):
    def name(self):
        return 'sub_Counter'


class sub_OrderedDict(OrderedDict):
    def name(self):
        return 'sub_OrderedDict'


SUB_CLASSES = [
    sub_object,
    sub_list,
    sub_dict,
    sub_set,
    sub_Counter,
    sub_OrderedDict,
]

COLL_CLASSES = [
    Counter,
    OrderedDict,
]

type_get = type.__getattribute__


async def verify_sub_class_can_instantiate(cls: type):

    pair = DebugWorld.connected_pair(logging='ALL')

    async with pair:

        A, B = pair.pipes

        cls_a = cls
        sup_a = cls.__bases__[0]
        cls_b = await B(cls_a)
        obj_a = cls_a()

        obj_b = await B(obj_a)
        assert obj_b.name() == obj_a.name()

        obj_b = cls_b()
        obj_b_cls = type(obj_b)
        assert obj_b_cls is not cls_a, f"{obj_b_cls} is {cls_a}"
        assert issubclass(obj_b_cls, sup_a), f"{obj_b_cls} is not subclass of {cls_a}"
        assert obj_b_cls is cls_b, f"{obj_b_cls} is not {cls_b}"

        obj_a = await A(obj_b)
        assert type(obj_a) is cls_a
        assert type(obj_a) is not cls_b

    return pair


async def verify_coll_class_can_instantiate(cls: type):

    pair = DebugWorld.connected_pair(logging='ALL')

    async with pair:

        A, B = pair.pipes

        cls_a = cls
        cls_b = await B(cls_a)
        obj_a = cls_a()

        obj_b = cls_b()
        obj_b_cls = type(obj_b)
        assert obj_b_cls is cls_a, f"{obj_b_cls} is not {cls_a}"

        obj_a2 = await A(obj_b)
        obj_a2_cls = type(obj_a2)
        assert obj_a2_cls is cls_a, f"{obj_b_cls} is not {cls_a}"
        assert obj_a == obj_a2

    return pair



def collect_collection_events():
    ...


async def verify_classes_can_instantiate():
    for cls in SUB_CLASSES:
        await verify_sub_class_can_instantiate(cls)
    for cls in COLL_CLASSES:
        await verify_coll_class_can_instantiate(cls)


if __name__ == '__main__':
    asyncio.run(verify_classes_can_instantiate())



# fmt: off

import sys
from collections.abc import Callable
from time import time_ns
from typing import ClassVar, NoReturn, Self

from ..color import WARN
from ..debug import fmt_exception

from .log_fns import *

__all__ = [
    'LogContext',
    'CancelMe',
    'is_user_exception',
    'no_log_ctx'
]

###############################################################################

class LogContext:
    LOGGING: ClassVar[bool] = True
    NULL: ClassVar['LogContext']
    PRINT_SLOW: ClassVar[float | int] = 30

    label: str
    title: tuple
    log: Callable[..., None]

    def __init__(self, label: str | tuple[str, str], *title) -> None:
        self.title = title or ()
        self.label = label
        self.log = log_fn(label) if self.LOGGING else no_log_fn

    def __bool__(self) -> bool:
        return True

    def __enter__(self) -> Self:
        self.start_ns = time_ns()
        self.log('(', *self.title, ')')
        return self

    def __call__(self, *args) -> None:
        if self.LOGGING:
            self.log(*args)

    def print(self, *args) -> None:
        log1(self.label, *args)

    def warn(self, *args) -> None:
        self.log(WARN('WARNING:'), *args)

    def vars(self, **kwargs) -> None:
        from ..print import f_multiline_dict
        self.log(f_multiline_dict(kwargs))

    def cancel_outer(self) -> NoReturn:
        raise CancelMe()

    def print_exception(self, exc: BaseException) -> None:
        f_exception = fmt_exception(exc)
        self.print(f_exception)

    def info(self, *args) -> None:
        self.log(*args)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        nanos = time_ns() - self.start_ns
        secs = nanos / 1e9

        # if secs > LogContext.PRINT_SLOW:
        #   print(f"PRINT_SLOW: {secs:.4f}s TAKEN BY", self.label, file=STDERR)

        if not self.LOGGING:
            return

        after = f'after {secs:.2f}s' if secs > 0.01 else ''
        if exc_type:
            exc_name = exc_type.__name__
            # virtual coroutines methods can return this, don't be noisy with logs
            # if exc_type in UNLOGGED_EXCEPTIONS:
            #     self.log(f'saw {exc_name}')
            #     return
            if 'Cancelled' in exc_name:
                self.log('was cancelled', after)
            elif exc_name == 'WarpShutdown':
                self.log('shutting down due to WarpShutdown', after)
            elif exc_type is CancelMe:
                self.log('chose to cancel; raising CancelledError', after)
                import asyncio

                raise asyncio.CancelledError(f'{self.label} chose to cancel')
            else:
                self.log('experienced exception', exc_name, after, '\n', exc_val)
        else:
            self.log('exited', after) if after else None

###############################################################################

class NoLogContext:

    def __bool__(self) -> bool:
        return False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def print(self, *args) -> None: ...
    def log(self, *args) -> None: ...
    def info(self, *args) -> None: ...
    def warn(self, *args) -> None: ...
    def vars(self, **kwargs) -> None: ...
    def cancel_outer(self) -> NoReturn: ...
    def print_exception(self, exc: BaseException) -> None: ...

no_log_ctx = NoLogContext()

###############################################################################

def is_user_exception(exc: Exception) -> bool:
    if not isinstance(exc, BaseException):
        return False
    exc_name = type(exc).__name__
    if exc_name in ('WarpShutdown', 'CancelMe', 'CancelledError'):
        return False
    if exc_name.startswith('Warp'):
        return False
    return True

###############################################################################

UNLOGGED_EXCEPTIONS = (
    RuntimeError,
    StopIteration,
    AttributeError,
    KeyError,
    ValueError,
    TypeError,
    NameError,
)

STDOUT = sys.stdout
STDERR = sys.stderr
EXIT_BANNER = '█' * 80

###############################################################################

class CancelMe(BaseException):
    pass



# fmt: off

from collections.abc import Callable
from dataclasses import dataclass, field

__all__ = [
    'ReplCallableInfo',
]

################################################################################

@dataclass
class ReplCallableInfo:
    """
    Represents information about a callable value in the repl.

    warp will automatically JSON-serialize instances within REPL messages
    since they operate with `fmt=JSON`.
    """

    fun_name:     str            = ''
    fun_qualname: str            = ''
    arg_names:    list[str]      = field(default_factory=list)
    arg_annos:    dict[str, str] = field(default_factory=dict)
    ret_anno:     str | None     = None
    doc_str:      str | None     = None
    is_async:     bool           = False
    fun_stub:     str | None     = None
    args_stub:    str | None     = None

    def set_from_function(self, clb: Callable, /):

        if not callable(clb):
            return

        from ..cpython.inspect import resolve_callable
        from ..cpython.function import func_sig_info
        from ..core.anno import anno_str

        fun, skip = resolve_callable(clb)
        sig_info = func_sig_info(fun)
        arg_names = sig_info.all_arg_names()
        arg_names = arg_names[skip:]

        def get[T](attr: str, default: T) -> T:
            value = getattr(clb, attr, None) or getattr(fun, attr, None)
            return value if type(value) is type(default) else default

        self.fun_name = fun_name = get('__name__', '')
        self.fun_qualname = get('__qualname__', '') or fun_name
        self.doc_str = get('__doc__', '') or None
        annos = get('__annotations__', {}).copy()

        self.arg_names = arg_names
        self.arg_annos = {k: anno_str(annos[k]) if k in annos else 'Any' for k in arg_names}
        self.ret_anno = anno_str(annos['return']) if 'return' in annos else 'Any'
        self.is_async = sig_info.is_async

    def __debug_info_str__(self) -> str:
        return f'name={self.fun_name!r} args={self.arg_names!r} ret={self.ret_anno!r}'

    def signature_str(self) -> str:
        f_args = ', '.join(f'{k}: {v}' for k, v in self.arg_annos.items())
        return f'({f_args})'



from typing import ClassVar

__all__ = ['Unset', 'UNSET']


class Unset:
    inst: ClassVar['Unset | None'] = None
    __slots__ = ()

    def __new__(cls) -> 'Unset':
        if cls.inst is None:
            cls.inst = super().__new__(cls)
        return cls.inst

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return '<UNSET>'


UNSET = Unset()



# fmt: off

import time
import sys
from io import StringIO

from os import getenv
from datetime import datetime

from collections.abc import Callable
from typing import Literal
from types import FunctionType, ModuleType
from pprint import pprint as _pprint
from threading import current_thread
from sys import setrecursionlimit, getrecursionlimit
from re import compile as re_compile
from agentica_internal.cpython.frame import exception_location

from .log import write_out, write_err
from .fmt import f_object_id, f_datum
from .color import *
from .ansi.palettes import Medium as MEDIUM
from .debug import debug_fmt, colorize
from .result import Result

__all__ = [
    'oprint',
    'eprint',
    'rprint',
    'tprint',
    'nprint',
    'hprint',
    'pprint',
    'sprint',
    'colorize',
    'hdiv',
    'ddiv',
    'remote_print',
    'local_print',
    'print_unclosed_error',
    'print_asyncio_stacks',
    'print_current_stack',
    'print_current_frame',
    'print_all_tasks',
    'print_all_threads',
    'print_exception',
    'print_traceback',
    'print_warning',
    'timestamp',
    'set_timestamp_fmt',
    'RESET',
    'WARN',
    'ERROR',
    'WHITE',
    'RED',
    'YELLOW',
    'GREEN',
    'BLUE',
    'MAGENTA',
    'CYAN',
    'BOLD',
    'ITALIC',
    'UNDER',
    'DIM',
    'MEDIUM'
]

####################################################################################################

def oprint(*args, sep: str = ' ', start: str = '', end='\n', file=None):
    """Print to the _original_ system stdout."""
    text = _sprint(args, sep, start, end, str)
    write_out(text)

def eprint(*args, sep: str = ' ', start: str = '', end='\n', file=None):
    """Print to the _original_ system stderr."""
    text = _sprint(args, sep, start, end, str)
    write_err(text)

def rprint(*args, sep: str = ' ', start: str = '', end='\n', err: bool = False):
    """Prints to either original stdout or stderr."""
    text = _sprint(args, sep, start, end, str)
    write_err(text) if err else write_out(text)

def sprint(*args, sep: str = ' ', start='', end=''):
    """Prints to a string with same semantics as builtin print."""
    return _sprint(args, sep, start, end, str)

####################################################################################################

def _sprint(args: tuple, sep: str, start: str, end: str, str_fn: Callable[[object], str]) -> str:
    start = start if type(start) is str else ''
    end = end if type(end) is str else '\n'
    sep = sep if type(sep) is str else ' '
    if not args:
        return start + end
    strs = [start]
    add = strs.append
    for i, arg in enumerate(args):
        add(sep) if i else None
        add(str_fn(arg))
    add(end)
    return cat(strs)

cat = ''.join

####################################################################################################

def remote_print(*args, sep: str = ' ') -> None:
    _remote_print(RP_GUTTER, args, sep)

def local_print(*args, sep: str = ' ') -> None:
    _remote_print(LP_GUTTER, args, sep)

def _remote_print_str(args: tuple, sep: str = ' ') -> str:
    if len(args) == 1 and type(args[0]) is str:
        return args[0]
    return _sprint(args, sep, '', '', f_object_id)

def _remote_print(gutter: str, args: tuple, sep: str) -> None:
    text = _remote_print_str(args, sep)
    text = add_gutter(gutter, text)
    write_out(text + '\n\n')

def add_gutter(gutter: str, text: str) -> str:
    return cat(gutter + s for s in text.splitlines(True))

RP_GUTTER = "\033[44m RP \033[49m "
LP_GUTTER = "\033[43m LP \033[49m "

####################################################################################################

def pprint(*args, err: bool = False):
    """Pretty prints, using rich if possible."""
    try:
        if 'rich' in sys.modules:
            import rich
            str_io = StringIO()
            rich.print(*args, file=str_io)
            text = str_io.getvalue()
            write_err(text) if err else write_out(text)
            return
    except:
        pass
    str_io = StringIO()
    _pprint(*args, stream=str_io, compact=False, sort_dicts=False, depth=5)
    text = str_io.getvalue()
    write_err(text) if err else write_out(text)

####################################################################################################

START = time.time()

_NOW_FMT: str | None = None

def set_timestamp_fmt(fmt: Literal[None, 'NONE', 'MS', 'SHORT', 'FULL']):
    global _NOW_FMT
    if fmt is None or fmt == 'NONE':
        _NOW_FMT = None
    elif fmt == 'MS':
        _NOW_FMT = '$'
    elif fmt == 'SHORT':
        _NOW_FMT = '@'
    elif fmt == 'FULL':
        _NOW_FMT = FULL_FORMAT + ': '
    else:
        raise ValueError(f'Unknown timestamp format {fmt!r}')

FULL_FORMAT = "%Y-%m-%d %H:%M:%S"

def timestamp() -> str:
    if _NOW_FMT is None:
        return ''
    elif _NOW_FMT == '$':
        millis = round((time.time() - START) * 1000)
        return f'{millis:<10} '
    elif _NOW_FMT == '@':
        now = datetime.now()
        millis = now.microsecond // 1000
        time_str = f'{now.minute:02d}:{now.second:02d}:{millis:03d}'
        return DIM(time_str) + ' '
    elif isinstance(_NOW_FMT, str):
        return datetime.now().strftime(_NOW_FMT)
    return ''

DEFAULT_FORMAT: Literal[None, 'MS', 'SHORT', 'FULL'] = None
set_timestamp_fmt(getenv('LOGGING_TIMESTAMP_FORMAT', DEFAULT_FORMAT)) # type: ignore

####################################################################################################

def tprint(*args, sep: str = ' ', end: str = '\n', truncate: bool = True, err: bool = False) -> None:
    """Print with a timestamp, thread name, and truncating long arguments."""
    limit = getrecursionlimit()
    text = ''
    try:
        setrecursionlimit(1000)
        thread_name = current_thread().name
        if truncate and args:
            args = trunc_args(args, len(thread_name)+1)
        thread_name = colorize(thread_name, 20)
        f_thread = f'\n{timestamp()}{thread_name} '
        text = _sprint(args, sep, f_thread, end, str)
    except Exception as exc:
        f_types = ' '.join(type(arg).__name__ for arg in args)
        f_location = exception_location(exc)
        text = f'\n\ntprint error: {type(exc).__name__}\nloc = {f_location}\nargs = {f_types}\n\n'
    finally:
        setrecursionlimit(limit)
    write_err(text) if err else write_out(text)

####################################################################################################

# TODO: rename `truncate` to `format`
def nprint(*args, sep: str = ' ', end: str = '\n', truncate: bool = True, err: bool = False):
    """Like tprint but without the thread name and timestamp."""
    limit = getrecursionlimit()
    text = ''
    try:
        setrecursionlimit(1000)
        if truncate and args:
            args = trunc_args(args, 0)
        text = _sprint(args, sep, '', end, str)
    except Exception as exc:
        f_types = ' '.join(type(arg).__name__ for arg in args)
        f_location = exception_location(exc)
        text = f'\n\nnprint error: {type(exc).__name__}\nloc = {f_location}\nargs = {f_types}\n\n'
    finally:
        setrecursionlimit(limit)
    write_err(text) if err else write_out(text)

####################################################################################################

def hprint(*args, sep: str = ' ', end: str = '\n'):
    """Print with a horizontal line above."""
    text = _sprint(args, sep, HDIV, end, str)
    write_out(text)

####################################################################################################

def hdiv():
    """Print a horizontal line."""
    write_out(HDIV)

def ddiv():
    """Print a double horizontal line."""
    write_out(DDIV)

####################################################################################################

def print_unclosed_error(obj):
    name = f_trunc(obj)
    text = f'{BANNER}{name} WAS GARBAGE COLLECTED WITHOUT BEING CLOSED\n{BANNER}'
    write_err(text)

####################################################################################################

def f_trunc(arg: object) -> str:

    cls = type(arg)

    if cls is bytes:
        s = f_bytes(arg)  # type: ignore
        return Rgb.hash_bytes(arg)(s)

    if cls is str and len(arg) > 256 and len(pure := strip_ansi(arg)) > 256:
        if 'Traceback (most recent call last)' in arg:
            return arg
        if '\n\n' in arg:
            return arg
        if arg.startswith("'''") and arg.endswith("'''"):
            return arg
        if ' := ' in arg:
            return arg
        return pure[:256] + "…"

    if cls is str:
        if arg.startswith('<') and arg.endswith('>'):
            return colorize(arg)
        if len(arg) == 36 and bool(UUID_RE.fullmatch(arg)):
            return colorize(f_uuid(arg))
        return arg

    if cls in (int, bool, float):
        return f_datum(arg)

    if cls is Result:
        if arg.is_ok:
            return f'Result.ok({f_object_id(arg.value)})'
        return repr(arg)

    kind: str = ''
    name: str = ''
    virt: bool = False

    if cls is type or isinstance(arg, type):
        kind = 'class'
        name = arg.__name__
        virt = is_virt_cls(arg)
    elif cls is FunctionType:
        kind = 'function'
        name = arg.__name__
        virt = is_virt_obj(arg)
    elif cls is ModuleType:
        name = arg.__name__
        kind = 'module'
        virt = is_virt_obj(arg)

    if name:
        if virt:
            kind = 'virtual ' + kind
        return colorize(f'<{kind} {name!r} @ {id(arg):x}>')

    if issubclass(cls, BaseException):
        return f_exception(arg)

    name = cls.__name__
    if is_virt_obj(arg):
        return colorize(f'<virtual {name!r} object @ {id(arg):x}>')
    elif is_virt_cls(cls):
        name = f'real object of virtual class {cls.__name__!r}'
        return colorize(f'<real object of virtual class {name!r} @ {id(arg):x}>')

    try:
        type_get(cls, 'short_str')
        f_short = arg.short_str()
        if type(f_short) is str:
            return colorize(f_short)
    except:
        pass

    if cls in (tuple, list, set, frozenset) and len(arg) <= 4:
        f_args = ', '.join(map(f_trunc, arg))
        if cls is tuple:
            return '(' + f_args + ')'
        if cls is frozenset:
            return 'frozenset(' + f_args + ')'
        if cls is list:
            return '[' + f_args + ']'
        if cls is set:
            return '{' + f_args + '}'

    if cls is dict and len(arg) <= 6:
        keys = tuple(arg.keys())
        if all(type(k) is str for k in keys):
            return 'dict(' + ', '.join(f'{k}=...' for k in keys) + ')'

    try:
        type_get(cls, '__short_str__')
        f_short = arg.__short_str__()
        if type(f_short) is str:
            return colorize(f_short)
    except:
        pass

    f_arg = f_object_id(arg)
    f_col = Rgb.id_obj(arg)
    return f_col.hi(f_arg)


def is_virt_cls(cls: type) -> bool:
    try:
        type_get(cls, '___vhdl___')
        return True
    except:
        return False

def is_virt_obj(obj: object) -> bool:
    try:
        obj_get(obj, '___vhdl___')
        return True
    except:
        return False


obj_get = object.__getattribute__
type_get = type.__getattribute__

####################################################################################################

def trunc_args(args: tuple[str, ...], margin: int) -> tuple[str, ...]:
    if not args:
        return ()
    *most, last = args
    # this handles the case where there is a single non-colorized final arg
    # that is a record, which we will print on multiple lines using 'k = v' syntax...
    if all(type(arg) is str and arg.startswith(ANSI_PREFIX) for arg in most):
        if isinstance(last, dict) and len(last) and all(type(k) is str for k in last.keys()):
            return *most, f_multiline_dict(last, margin)
    return tuple(map(f_trunc, args))

ANSI_PREFIX = "\033"

####################################################################################################

def f_multiline_dict(dct: dict, margin: int = -1) -> str:
    if not dct:
        return '{}'
    if margin == -1:
        margin = len(current_thread().name) + 1
    keys = list(dct.keys())
    wide = max(map(len, keys))
    f_margin = ' ' * margin
    result = ''.join(f'\n{f_margin}{k.ljust(wide)} := {f_trunc(v)}' for k, v in dct.items())
    return result

####################################################################################################

def f_uuid(s: str) -> str:
    return colorize(s.rsplit('-', 1)[-1])

####################################################################################################

def f_msgpack(m: bytes) -> str:
    from itertools import takewhile
    if m[1:].startswith(b'\xa3msg'):
        msg_kind = ''.join(map(chr, takewhile(lambda x: 32 < x < 128, m[6:32])))
        return f"<{msg_kind!r} msgpack bytes>"
    else:
        return "<unknown msgpack bytes>"

####################################################################################################

def f_bytes(b: bytes) -> str:
    if len(b) > 8 and 0x80 <= b[0] <= 0x8F:
        return f_msgpack(b)
    if len(b) > 48:
        s = repr(b[:48])
        return s[:48] + "…" + s[0]
    return repr(b)

####################################################################################################

def f_exception(exc: BaseException):
    import traceback
    lines = traceback.format_exception(exc)
    if location := exception_location(exc):
        lines.insert(0, location)
    f_exc = '\n'.join(lines).replace('\n', INDENT).rstrip(INDENT)
    return INDENT + f_exc

####################################################################################################

panic                = debug_fmt.panic
print_current_stack  = debug_fmt.print_current_stack
print_current_frame  = debug_fmt.print_current_frame
print_all_tasks      = debug_fmt.print_all_tasks
print_all_threads    = debug_fmt.print_all_threads
print_exception      = debug_fmt.print_exception
print_traceback      = debug_fmt.print_traceback
print_warning        = debug_fmt.print_warning

###############################################################################


def print_asyncio_stacks(delay: int = 0):
    if delay > 0:
        return print_asyncio_stacks_in_n_seconds(delay)

    write_out(STACKS_TEXT)
    try:
        print_all_tasks()
    except Exception as exc:
        oprint("Error printing asyncio stacks:", type(exc).__name__)

    write_out(THREADS_TEXT)

    try:
        print_all_threads()
        oprint('', BDIV, '', sep='\n')
    except Exception as exc:
        oprint("Error printing threads:", type(exc).__name__)


def print_asyncio_stacks_in_n_seconds(delay: int):
    import signal

    def debug_handler(signum, frame):
        del signum
        del frame
        print_asyncio_stacks()

    signal.signal(signal.SIGALRM, debug_handler)
    signal.alarm(delay)


###############################################################################

HDIV = BLUE('─' * 80) + '\n'
DDIV = RED('═' * 80) + '\n'
BDIV = RED('█' * 80) + '\n'

INDENT = '\n┃ '

BANNER = (("*" * 75) + "\n") * 10

UUID_RE = re_compile(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')

STACKS_TEXT = f'\n{BDIV}ASYNCIO STACKS\n\n'
THREADS_TEXT = f'\n{BDIV}THREADS\n\n'



import asyncio

from agentica_internal.testing.examples.classes.generic import GenericClass
from agentica_internal.warpc.worlds.all import DebugWorld


async def run():
    async with DebugWorld.connected_pair(logging='ENCR+DECR+SEND+RECV') as B:
        cls = GenericClass

        cls_b = await B(cls)
        print(type(cls), type(cls_b))
        print(cls, cls_b)


asyncio.run(run(), debug=True)



# fmt: off

import struct
from collections.abc import Callable
from types import ModuleType

from xxhash import xxh64, xxh64_intdigest

# from .singletons import INDETERMINATE, MISSING

__all__ = [
    'stable_hash',
    '__stable_hash__',
    'raw_str_hash',
    'raw_strs_hash',
    'raw_bytes_hash',
    'raw_hashes_hash',
    'raw_type_hash',
    'raw_int_hash',
    'combine_hashes',
    'HASH_MASK',
]


HASH_MASK = (1 << 63) - 1

type Biter = Callable[[bytes | bytearray | str], None]


def raw_bytes_hash(b: bytes) -> int:
    return xxh64_intdigest(b) & HASH_MASK


def raw_str_hash(string: str) -> int:
    return xxh64_intdigest(string) & HASH_MASK


def raw_strs_hash(strings: tuple[str, ...] | list[str]) -> int:
    hasher = xxh64()
    bite = hasher.update
    bite(len(strings).to_bytes(16))
    bite('\x01'.join(strings))
    bite(b'\xff')
    return hasher.intdigest() & HASH_MASK


def raw_hashes_hash(hashes: list[int]) -> int:
    hasher = xxh64()
    bite = hasher.update
    bite(len(hashes).to_bytes(16))
    bite(b'\x01'.join((h & HASH_MASK).to_bytes(16) for h in hashes))
    bite(b'\xff')
    return hasher.intdigest() & HASH_MASK


def raw_type_hash(cls: type) -> int:
    cls_name = cls.__module__ + '.' + cls.__qualname__
    return raw_bytes_hash(b'type\0' + cls_name.encode() + b'\1')


def __stable_hash__(obj) -> int:
    if hasattr(obj, '__hash_value__'):
        return obj.__hash_value__
    hasher = xxh64()
    update = hasher.update
    cls = obj.__class__
    name = cls.__module__ + '.' + cls.__qualname__
    update(b'object\0' + name.encode())
    fields: tuple[str, ...] = obj.__hash_args__
    # print(name, len(fields))
    update(len(fields).to_bytes(8))
    for field in fields:
        update(b'attr\0')
        update(field)
        update(b'value\0')
        for val in _bites(getattr(obj, field, MISSING)):
            update(val)
    hash_value = hasher.intdigest() & HASH_TRIM
    setattr(obj, '__hash_value__', hash_value)
    return hash_value


def stable_hash(obj: object) -> int:
    if hasattr(obj, '__hash_value__'):
        return obj.__hash_value__  # type: ignore
    if hasattr(obj.__class__, '__hash_args__'):
        return __stable_hash__(obj)
    hasher = xxh64()
    update = hasher.update
    for bite in _bites(obj):
        update(bite)
    return hasher.intdigest() & HASH_TRIM


def _bites(obj: object):
    if isinstance(obj, int):
        if obj >= 0:
            yield b'int\0'
            yield obj.to_bytes(16)
        else:
            yield b'negint\0'
            yield abs(obj).to_bytes(16)
    elif isinstance(obj, float):
        yield b'float\0'
        yield struct.pack('d', obj)
    elif isinstance(obj, bool):
        yield b'True\0' if obj else 'False\0'
    elif obj is None:
        yield b'None\0'
    elif isinstance(obj, str):
        yield b'str\0'
        yield len(obj).to_bytes(16)
        yield obj
    elif isinstance(obj, type):
        cls_name = obj.__module__ + '.' + obj.__qualname__
        yield b'type\0'
        yield len(cls_name).to_bytes(16)
        yield cls_name.encode()
    elif hasattr(obj, '__hash_value__'):
        yield obj.__hash_value__.to_bytes(16)  # type: ignore
    elif isinstance(obj, bytes):
        yield b'bytes\0'
        yield len(obj).to_bytes(16)
        yield obj
    elif isinstance(obj, tuple):
        yield b'tuple\0'
        yield len(obj).to_bytes(16)
        for el in obj:
            yield from _bites(el)
    elif isinstance(obj, list):
        yield b'list\0'
        yield len(obj).to_bytes(16)
        for el in obj:
            yield from _bites(el)
    elif isinstance(obj, set):
        yield b'set\0'
        yield len(obj).to_bytes(16)
        for el in obj:
            yield from _bites(el)
    elif isinstance(obj, dict):
        yield b'dict\0'
        yield len(obj).to_bytes(16)
        for k, v in obj.items():
            yield from _bites(k)
            yield from _bites(v)
    elif isinstance(obj, ModuleType):
        yield b'module\0'
        yield obj.__name__ or ''
        yield obj.__package__ or ''
    elif hasattr(obj.__class__, '__hash_args__'):
        yield __stable_hash__(obj).to_bytes(16)  # type: ignore
    else:
        raise ValueError(f"cannot hash value of type {obj.__class__.__name__}")


HASH_TRIM = (1 << 57) - 1  # so 23 * won't overflow


def combine_hashes(a: int, b: int) -> int:
    return (17 * ((a & HASH_TRIM) | 1) + 23 * ((b & HASH_TRIM) | 2)) & HASH_MASK


def raw_int_hash(i: int) -> int:
    n = 0xCBF29CE484222325 ^ i
    n = n ^ (n >> 33)
    n = n * 0xFF51AFD7ED558CCD
    n = n ^ (n >> 33)
    n = n * 0xC4CEB9FE1A85EC53
    n = n ^ (n >> 33)
    return n & HASH_MASK



# fmt: off

import builtins
import sys
import threading
from collections.abc import Callable, Iterable
from contextlib import contextmanager
from typing import NamedTuple, TextIO

__all__ = [
    'SystemHooks',
    'original_print',
    'original_stdout',
    'original_stderr',
    'original_stdin',
]

################################################################################

type PrintFn = Callable[..., None]

# Store the original system values before any REPL manipulates them
_original_stdin = sys.stdin
_original_stdout = sys.stdout
_original_stderr = sys.stderr
_original_print = builtins.print

# Export these for use by other modules that need the true originals
original_stdin = _original_stdin
original_stdout = _original_stdout
original_stderr = _original_stderr
original_print = _original_print

# Thread-local storage for per-thread hooks
_thread_local = threading.local()


def _get_thread_hooks() -> 'SystemHooks | None':
    """Get the hooks for the current thread, or None if not set."""
    return getattr(_thread_local, 'hooks', None)


def _set_thread_hooks(hooks: 'SystemHooks | None') -> None:
    """Set the hooks for the current thread."""
    _thread_local.hooks = hooks


class ThreadLocalTextIO(TextIO):
    """
    A delegating wrapper around TextIO that looks up the actual stream
    from thread-local storage. Falls back to the original stream if no
    thread-local hook is set.
    """
    def __init__(self, attr_name: str, original: TextIO):
        self._attr_name = attr_name
        self._original = original

    def _get_stream(self) -> TextIO:
        hooks = _get_thread_hooks()
        if hooks is not None:
            return getattr(hooks, self._attr_name)
        return self._original

    def write(self, s: str) -> int:
        return self._get_stream().write(s)

    def writelines(self, lines: Iterable[str]) -> None:
        return self._get_stream().writelines(lines)

    def read(self, size: int = -1) -> str:
        return self._get_stream().read(size)

    def readline(self, size: int = -1) -> str:
        return self._get_stream().readline(size)

    def readlines(self, hint: int = -1) -> list[str]:
        return self._get_stream().readlines(hint)

    def flush(self) -> None:
        return self._get_stream().flush()

    def close(self) -> None:
        # Don't close the underlying stream
        pass

    def isatty(self) -> bool:
        return self._get_stream().isatty()

    def readable(self) -> bool:
        return self._get_stream().readable()

    def writable(self) -> bool:
        return self._get_stream().writable()

    def seekable(self) -> bool:
        return self._get_stream().seekable()

    def fileno(self) -> int:
        return self._get_stream().fileno()

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._get_stream().seek(offset, whence)

    def tell(self) -> int:
        return self._get_stream().tell()

    def truncate(self, size: int | None = None) -> int:
        return self._get_stream().truncate(size)

    @property
    def encoding(self) -> str:
        return self._get_stream().encoding

    @property
    def errors(self) -> str | None:
        return self._get_stream().errors

    @property
    def line_buffering(self) -> bool:
        return bool(self._get_stream().line_buffering)

    @property
    def newlines(self):
        return self._get_stream().newlines

    @property
    def buffer(self):
        return self._get_stream().buffer

    @property
    def mode(self) -> str:
        return self._get_stream().mode

    @property
    def name(self):
        return self._get_stream().name

    @property
    def closed(self) -> bool:
        return self._get_stream().closed

    def __iter__(self):
        return iter(self._get_stream())

    def __next__(self):
        return next(self._get_stream())

    def __enter__(self):
        return self._get_stream().__enter__()

    def __exit__(self, *args):
        return self._get_stream().__exit__(*args)


def _thread_local_print(*args, sep: str = ' ', end: str = '\n', file=None, flush: bool = False) -> None:
    """
    Thread-local print function that delegates to the per-thread print hook
    if one is set, otherwise uses the original print.
    """
    hooks = _get_thread_hooks()
    if hooks is not None:
        # Use the thread-local print function
        hooks.print_fn(*args, sep=sep, end=end, file=file, flush=flush)
    else:
        # Fall back to original print
        _original_print(*args, sep=sep, end=end, file=file, flush=flush)


# Install thread-local wrappers once at import time
# This ensures that sys.stdout etc. always delegate to thread-local hooks
_delegating_stdin = ThreadLocalTextIO('stdin', _original_stdin)
_delegating_stdout = ThreadLocalTextIO('stdout', _original_stdout)
_delegating_stderr = ThreadLocalTextIO('stderr', _original_stderr)

sys.stdin = _delegating_stdin  # type: ignore
sys.stdout = _delegating_stdout  # type: ignore
sys.stderr = _delegating_stderr  # type: ignore
builtins.print = _thread_local_print

################################################################################


class SystemHooks(NamedTuple):
    """
    Represents the changes of global Python state that must be made during
    a REPL execution.

    Make a desired state with `state = SystemHooks.make(...)`.

    Temporarily apply it with `with state.applied(): ...`

    Permanently apply it with `state.set()`.

    Get the current system state with `SystemHooks.get()`.
    
    NOTE: Hooks are now thread-local. Each thread has its own set of hooks,
    preventing concurrent sandboxes from interfering with each other's stdout.
    """

    stdin:           TextIO
    stdout:          TextIO
    stderr:          TextIO
    print_fn:        PrintFn
    recursion_limit: int

    @staticmethod
    def make(
            stdin: TextIO,
            stdout: TextIO,
            stderr: TextIO, *,
            print_fn: PrintFn = builtins.print,
            recursion_limit: int = 0) -> 'SystemHooks':

        return SystemHooks(
            stdin, stdout, stderr,
            print_fn,
            recursion_limit,
        )

    def set(self) -> None:
        """
        Set the hooks for the current thread.
        
        This is now thread-local, so concurrent threads can have different hooks
        without interfering with each other.
        """
        _set_thread_hooks(self)
        # Ensure sys.stdin/stdout/stderr are always our delegating wrappers
        # (user code might have reassigned them)
        sys.stdin = _delegating_stdin
        sys.stdout = _delegating_stdout
        sys.stderr = _delegating_stderr
        builtins.print = _thread_local_print
        # Recursion limit is still process-global, but that's generally safe
        if self.recursion_limit:
            sys.setrecursionlimit(self.recursion_limit)

    @staticmethod
    def get() -> 'SystemHooks':
        """
        Get the current hooks for this thread.
        
        If no thread-local hooks are set, returns the original system values.
        """
        hooks = _get_thread_hooks()
        if hooks is not None:
            return hooks
        rec_limit = sys.getrecursionlimit()
        return SystemHooks(
            _original_stdin, _original_stdout, _original_stderr,
            _original_print,
            rec_limit,
        )

    @staticmethod
    def clear() -> None:
        """Clear the thread-local hooks, reverting to original system streams."""
        _set_thread_hooks(None)
        # Ensure sys.stdin/stdout/stderr are always our delegating wrappers
        # (user code might have reassigned them)
        sys.stdin = _delegating_stdin
        sys.stdout = _delegating_stdout
        sys.stderr = _delegating_stderr
        builtins.print = _thread_local_print

    @contextmanager
    def applied(self):
        old = SystemHooks.get()
        self.set()
        try:
            yield
        finally:
            if old == SystemHooks(
                _original_stdin, _original_stdout, _original_stderr,
                _original_print, old.recursion_limit
            ):
                # If old was the original state, clear instead
                SystemHooks.clear()
            else:
                old.set()



from .multiplex_protocol import *

__all__ = [
    "MultiplexErrorName",
    "MultiplexCancelMessage",
    "MultiplexClientInstanceMessage",
    "MultiplexClientMessage",
    "MultiplexDataMessage",
    "MultiplexErrorMessage",
    "MultiplexInvocationEventMessage",
    "MultiplexInvocationEventType",
    "MultiplexInvokeMessage",
    "MultiplexMessage",
    "MultiplexNewIIDResponse",
    "MultiplexServerInstanceMessage",
    "MultiplexServerMessage",
    "multiplex_from_json",
    "multiplex_to_json",
]



import pytest
from agentica_internal.warpc.worlds.debug_world import DebugWorld


@pytest.mark.asyncio
async def test_resource_pre_hooks():
    class MyClass:
        a: int
        b: str
        c: bool = True

        def __init__(self, a: int, b: str):
            self.a = a
            self.b = b

    my_obj = MyClass(2, 'hello')

    pair = DebugWorld.connected_pair()
    async with pair as B:
        my_obj_b = await B(my_obj)
        assert my_obj_b.a == 2
        assert my_obj_b.b == 'hello'
        assert my_obj_b.c is True

        assert {'a', 'b', 'c'} & set(dir(my_obj_b)) == {'a', 'b', 'c'}
        assert set(dir(my_obj_b)) == set(dir(MyClass) + ['a', 'b'])
        assert dir(my_obj_b) == dir(my_obj)

        assert my_obj_b.__dict__ == my_obj.__dict__ == {'a': 2, 'b': 'hello'}



# fmt: off

from dataclasses import dataclass, field
from typing import Literal

from ..warpc.predicates import *
from .repl_callable_info import ReplCallableInfo
from .repl_var_info import VarKind

__all__ = [
    'ReplResourcesInfo',
    'ReplReturnInfo',
    'ReplSystemInfo',
    'ReplSessionInfo',
    'ReplRole',
    'VALID_REPL_ROLES',
]


################################################################################

type ReplRole = Literal['function', 'agent']

VALID_REPL_ROLES = 'function', 'agent'

################################################################################

@dataclass
class ReplResourcesInfo:

    names:  tuple[str, ...]    = ()
    kinds:  dict[str, VarKind] = field(default_factory=dict)
    reprs:  dict[str, str]     = field(default_factory=dict)
    stub:   str                = ''

    def __bool__(self) -> bool:
        return bool(self.names) and bool(self.stub.strip())

    def clear(self):
        self.names = ()
        self.kinds.clear()
        self.reprs.clear()
        self.stub = ''

    def update(self, names, kinds, reprs, stub):
        self.names = tuple(dict.fromkeys(self.names) | dict.fromkeys(names))  # preserve order
        self.kinds.update(kinds)
        self.reprs.update(reprs)
        self.stub = stub or self.stub  # preserve the old stub if the new one is empty

    def __debug_info_str__(self) -> str:
        return f'names={self.names}'


################################################################################

@dataclass
class ReplSystemInfo:

    # python_version: str
    modules: tuple[str, ...] = ()

    def __debug_info_str__(self) -> str:
        return f'modules=<{len(self.modules)}>'


################################################################################

@dataclass
class ReplReturnInfo:

    type_str: str  = ''
    is_text:  bool = False
    is_none:  bool = False

    def __debug_info_str__(self) -> str:
        return f'type={self.type_str}'


################################################################################

@dataclass
class ReplSessionInfo:
    """
    Represents all the information needed to do prompting at once,
    so that we do not need to make repeated calls to the AgentRepl.

    This object is capable of accepting json-formatted updates via
    `.update`, and `AgentRepl.initialize()` will emit exactly the output
    needed to keep the `ReplSessionInfo` up-to-date with warp-driven changes.
    """

    role:       ReplRole
    returns:    ReplReturnInfo
    locals:     ReplResourcesInfo
    globals:    ReplResourcesInfo
    system:     ReplSystemInfo
    agentic_fn: ReplCallableInfo

    @staticmethod
    def empty() -> 'ReplSessionInfo':
        return ReplSessionInfo(
            role='agent',
            returns=ReplReturnInfo(),
            locals=ReplResourcesInfo(),
            globals=ReplResourcesInfo(),
            system=ReplSystemInfo(),
            agentic_fn=ReplCallableInfo(),
        )

    ############################################################################

    @property
    def is_function(self) -> bool:
        return self.role == 'function'

    @property
    def is_agent(self) -> bool:
        return self.role == 'agent'

    ############################################################################

    @property
    def return_type(self) -> str:
        return self.returns.type_str

    @property
    def is_returning_text(self) -> bool:
        return self.returns.is_text

    @property
    def is_returning_none(self) -> bool:
        return self.returns.is_none

    ############################################################################

    @property
    def has_arguments(self) -> bool:
        return bool(self.agentic_fn.arg_names)

    @property
    def function_name(self) -> str:
        return self.agentic_fn.fun_name

    @property
    def function_argument_names(self) -> list[str]:
        return self.agentic_fn.arg_names

    @property
    def function_argument_types(self) -> dict[str, str]:
        return self.agentic_fn.arg_annos

    @property
    def function_arguments_stub(self) -> str:
        return self.agentic_fn.args_stub

    @property
    def function_argument_signature(self) -> str:
        return self.agentic_fn.signature_str()

    @property
    def function_description(self) -> str:
        return self.agentic_fn.doc_str or ''

    @property
    def function_stub(self) -> str:
        return self.agentic_fn.fun_stub

    ############################################################################

    @property
    def available_modules(self) -> tuple[str, ...]:
        return self.system.modules

    @property
    def has_global_resources(self) -> bool:
        return bool(self.globals.stub)

    @property
    def has_local_resources(self) -> bool:
        return bool(self.locals.stub)

    @property
    def global_resources(self) -> tuple[str, ...]:
        return self.globals.names

    @property
    def local_resources(self) -> tuple[str, ...]:
        return self.locals.names

    @property
    def global_resources_stub(self) -> str:
        return self.globals.stub

    @property
    def local_resources_stub(self) -> str:
        return self.locals.stub

    @property
    def inputs_stub(self) -> str:
        return self.locals.stub

    @property
    def input_reprs(self) -> dict[str, str]:
        return self.locals.reprs

    ############################################################################

    def __template_vars__(self) -> set[str]:
        return SESSION_VARS

    def __debug_info_str__(self) -> str:
        return f'role={self.role} ret={self.returns.type_str}'

    ############################################################################

    def update(self, dct: dict[str, object]):
        self.__update_field(dct, 'returns',    ReplReturnInfo)
        self.__update_field(dct, 'locals',     ReplResourcesInfo)
        self.__update_field(dct, 'globals',    ReplResourcesInfo)
        self.__update_field(dct, 'system',     ReplSystemInfo)
        self.__update_field(dct, 'agentic_fn', ReplCallableInfo)
        self.__dict__.update(dct)

    def __update_field(self, dct: dict[str, object], name: str, cls: type):
        kwargs = dct.pop(name, None)
        if isinstance(kwargs, dict):
            setattr(self, name, cls(**kwargs))

################################################################################

SESSION_VARS = {
    'is_agentic_fn',
    'is_agent',
    # for agentic functions AND agents
    'task_description',
    'return_type',
    'is_returning_text',
    'is_returning_none',
    # for agentic functions
    'has_arguments',
    'function_name',
    'function_description',
    'function_argument_names',
    'function_argument_types',
    'function_argument_signature',
    'function_arguments_stub',
    'function_stub',
    # system and resource information
    'available_modules',
    'has_global_resources',
    'global_resources',
    'global_resources_stub',
    'has_local_resources',
    'local_resources',
    'local_resources_stub',
    # inputs
    'inputs_stub',
    'inputs_reprs',
}

################################################################################

def update_field(dct: dict[str, object], name: str, cls: type) -> None:
    kwargs = dct.pop(name)
    dct[name] = cls(**kwargs) if is_rec(kwargs) else cls()



import pytest
from agentica_internal.testing.warpc.virtual_warp_as_tests import *


@pytest.mark.asyncio
async def test_warp_as():
    await verify_warp_as()


@pytest.mark.asyncio
async def test_class_warp_as():
    await verify_class_warp_as()



# fmt: off

from typing import *
import os

from agentica_internal.testing import *
from agentica_internal.testing.examples import ANNOS, CLASSES
from agentica_internal.warpc.worlds.debug_world import DebugWorld


def dump_msg(v: Any):
    world = DebugWorld(world_id=9, logging=False)
    frame = world.root
    enc_context = frame.enc_context()
    with enc_context:
        msg = world.codec.enc_any(v)
    defs = enc_context.enc_context_defs()
    return msg.msgpack_str() + '\n' + '\n'.join(d.msgpack_str() for d in defs)


class MyThing:
    pass


MyThing.__module__ = 'my_module'

my_thing = MyThing()

CUSTOM = [
    [my_thing, my_thing, my_thing],
    os.stat_result([33188, 19003800, 16777233, 1, 501, 20, 3830, 1761816995, 1761590140, 1761590140]
),
]

file = FileTemplate(__file__, '{kind}.json', section='{name}')


def verify_msgpack_tests(pytest_request=None):
    objects = ANNOS + CLASSES + CUSTOM
    run_object_to_file_tests(
        dump_msg, objects, results_file=file, on_error='p', pytest_request=pytest_request
    )


if __name__ == '__main__':
    verify_msgpack_tests()



# fmt: off

from enum import Enum, Flag, IntEnum, IntFlag, StrEnum

from ..shared import *

__all__ = [
    'EnumOf3Ints',
    'EnumOf3Strs',
    'EnumOf3Mixed',
    'EnumOf3Flag',
    'EnumOf3IntFlag',
    'ENUM_CLASSES',
    'ENUM_INSTANCES',
]


class EnumOf3Ints(IntEnum):
    """EnumOf3Ints docstring"""

    a = _int1
    b = _int2
    c = _int3


class EnumOf3Strs(StrEnum):
    """EnumOf3Strs docstring"""

    a = _str1
    b = _str2
    c = _str3


class EnumOf3Mixed(Enum):
    """EnumOf3Mixed docstring"""

    a = _int1
    b = _str2
    c = _str3


class EnumOf3Flag(Flag):
    """EnumOf3Mixed docstring"""

    a = _int1
    b = _int3
    c = _int3


class EnumOf3IntFlag(IntFlag):
    """EnumOf3Mixed docstring"""

    a = _int1
    b = _int3
    c = _int3


ENUM_CLASSES: list[type] = [
    EnumOf3Ints,
    EnumOf3Strs,
    EnumOf3Mixed,
    EnumOf3Flag,
    EnumOf3IntFlag,
]

ENUM_INSTANCES: list[object] = [
    EnumOf3Ints.a,
    EnumOf3Strs.a,
    EnumOf3Mixed.a,
    EnumOf3Flag.a,
    # EnumOf3Flag.a | EnumOf3Flag.b,
    EnumOf3IntFlag.a,
    EnumOf3IntFlag.a | EnumOf3IntFlag.b,
]



# fmt: off

from typing import Callable

from .resource.virtual_function import describe_real_function, create_proxy_function

__all__ = [
    'create_magic_proxy_function'
]


################################################################################

def create_magic_proxy_function(src: Callable, dst: Callable):

    data = describe_real_function(src, process_defaults=False)

    return create_proxy_function(data, dst)



# fmt: off

from asyncio import AbstractEventLoop, Queue, Task, Future, TaskGroup, CancelledError  # noqa
from asyncio import to_thread, run_coroutine_threadsafe, get_running_loop, create_task  # noqa

from ..exceptions import raise_shutdown
# Queue.shutdown only arrived in 3.13, we emulate it for 3.12
from ...core.queues import Queue, QueueShutDown
from ...core.fmt import f_set

from ..request.vars import Vars

from .__ import *
from .base_world import *

__all__ = [
    'SDKWorld',
]

no_send_msg = cast(AsyncSendMsg, raise_shutdown)
no_recv_msg = cast(AsyncRecvMsg, raise_shutdown)

################################################################################

INVOCATION_MID = -1

class SDKWorld(World):
    """
    Handles virtualizing resources, servicing remote resource requests on
    the client side, and transmitting values via `channel_send` and `channel_recv`.

    `send_msg` and `recv_msg` will raise `WarpShutdown` unless an event
    loop task is running.
    """
    executing:  dict[MessageID, Task]  # locally executing request from remote
    awaiting:   dict[MessageID, Future[FramedResponseMsg]]
    futures:    dict[MessageID, Future]
    channel:    Queue[ChannelMsg]

    send_msg:   AsyncSendMsg
    recv_msg:   AsyncRecvMsg
    event_loop: AbstractEventLoop
    tasks:      list[Task]

    def __post_init__(self):
        super().__post_init__()
        self.executing = {}
        self.awaiting = {}
        self.futures = {}
        self.channel = Queue()
        self.task = None
        self.send_msg = no_send_msg
        self.tasks = []

    def clone(self) -> Self:
        cloned = super().clone()
        cloned.executing = {}
        cloned.awaiting = {}
        cloned.futures = {}
        cloned.channel = Queue()
        cloned.task = None
        cloned.inbox = Queue()
        cloned.outbox = Queue()
        cloned.tasks = []
        return cloned

    def __repr__(self) -> str:
        executing = f_set(f_id, self.executing.keys())
        awaiting = f_set(f_id, self.awaiting.keys())
        return f'SDKWorld(id={f_id(self.id)}, {executing=}, {awaiting=})'

    ############################################################################

    def to_payload(self, vars_: Rec[TermT]) -> bytes:
        vars_obj = Vars(vars_)
        vars_msg = vars_obj.encode(self.root, 'full')
        vars_data = vars_msg.to_msgpack()
        return vars_data

    ############################################################################

    def start_msg_loop(self,
                       send_bytes: AsyncSendBytes,
                       recv_bytes: AsyncRecvBytes,
                       future: Future | None = None) -> \
            Task:
        if not hasattr(self, 'event_loop'):
            self.event_loop = get_running_loop()

        log_send = bool(LOG_SEND)
        log_recv = bool(LOG_RECV)

        # rough idea here: a message loop is associated with ONE

        async def send_msg(msg: RPCMsg) -> None:
            try:
                self.on_message(Outgoing, self.tick, msg)
                msg_data = msg.to_msgpack()
                if log_send:
                    P.nprint(ICON_SEND, fmt_msgpack(msg_data, multiline=False))
                await send_bytes(msg_data)
            except RuntimeError:
                raise E.WarpShutdown()

        async def recv_msg() -> RPCMsg:
            msg_data = await recv_bytes()
            if log_recv:
                P.nprint(ICON_RECV, fmt_msgpack(msg_data, multiline=False))
            msg = RPCMsg.from_msgpack(msg_data)
            self.on_message(Incoming, self.tick, msg)
            return msg

        self.send_msg = send_msg

        async def msg_loop() -> None:
            try:
                while True:
                    rpc_msg = await recv_msg()
                    await self.handle_incoming_message(rpc_msg)
            except BaseException as exc:
                if future is not None:
                    if not future.done():
                        future.set_exception(exc)
                if type(exc) not in (CancelledError, E.WarpShutdown):
                    self.handle_msg_loop_err(exc)

        task_coro = msg_loop()
        task_name = f'{self.log_name}.msg_loop'
        task = self.event_loop.create_task(task_coro, name=task_name)
        self.tasks.append(task)
        return task

    def handle_msg_loop_err(self, exc: BaseException) -> None:
        self.log("handle_msg_loop_err", exc)

    # TODO: collapse this with new virtual futures
    def create_future(self, future_id: FutureID, future: Future | None = None) -> Future[Result]:
        with self.log_as("create_future", future_id) as ctx:
            loop = getattr(self, 'event_loop', None)
            if loop is None:
                self.event_loop = loop = get_running_loop()
            futures = self.futures
            previous = futures.get(future_id)
            if previous is None:
                if future is None:
                    ctx.log('created new future', future)
                    future = loop.create_future()
                else:
                    ctx.log('using provided future', future)
                self.futures[future_id] = future
            else:
                ctx.log('future already exists')
            return future

    ############################################################################

    def _new_future(self, future_id: FutureID, /) -> FutureT | None:
        future = new_hookable_future(self.event_loop)
        setattr(future, FUTURE_ID, future_id)
        return future

    def _await_future(self, future: FutureT, _: None, /) -> None:
        pass

    def _gather_future(self, future: FutureT, _: None, /) -> None:
        pass

    def send_event(self, msg: EventMsg) -> None:
        self.on_event(Outgoing, self.tick, msg)
        name = f'{self.log_name}.send_event'
        self.event_loop.create_task(self.send_msg(msg), name=name)

    ############################################################################

    # to satisfy WorldP

    def execute_outgoing_request_sync(self, msg: FramedRequestMsg) -> FramedResponseMsg:
        # we can't break the sync -> async barrier since the rest of agentica uses the
        # main event loop
        raise NotImplementedError('sync virtual requests not supported in this world')

    # NOTE: should this catch exceptions in the await response and turn these into RuntimeError?
    def execute_outgoing_request_coro(self, msg: FramedRequestMsg) -> Awaitable[FramedResponseMsg]:
        on_exit = self.on_frame(Outgoing, msg)
        loop = self.event_loop
        self.awaiting[msg.mid] = future = loop.create_future()

        async def await_response():
            await self.send_msg(msg)
            response = await future
            on_exit(response)
            return response

        await_response_coro = await_response()
        # this task will be cancelled when the future is cancelled, so we don't need to cancel it
        # ourselves
        name = f'{self.log_name}.await_response[{f_id(msg.mid)}]'
        loop.create_task(await_response_coro, name=name)
        return future

    def execute_outgoing_request_future(self, msg: FramedRequestMsg) -> Future[FramedResponseMsg]:
        loop = self.event_loop
        self.awaiting[msg.mid] = future = loop.create_future()
        name = f'{self.log_name}.outgoing_request[{f_id(msg.mid)}]'
        self.event_loop.create_task(self.send_msg(msg), name=name)
        return future

    ############################################################################

    def close(self) -> None:
        if self.closed:
            return
        self.closed = True
        self.__cancel()
        super().close()

    def __cancel(self) -> None:
        self.send_msg = no_send_msg
        self.recv_msg = no_recv_msg
        for task in self.tasks:
            if not task.done():
                task.cancel()
        for task in self.executing.values():
            task.cancel()
        self.executing.clear()
        for future in self.awaiting.values():
            future.cancel()
        self.awaiting.clear()
        for future in self.futures.values():
            self.unregister_future(future)
        self.futures.clear()
        self.channel.shutdown()  # type: ignore

    def needs_close(self) -> bool:
        if self.closed:
            return False
        if not self.executing and not self.awaiting:
            return False
        for executing in self.executing.values():
            if not executing.done():
                return True
        for task in self.tasks:
            if not task.done():
                return True
        self.executing.clear()
        self.tasks.clear()
        return False

    def __del__(self) -> None:
        if self.needs_close():
            P.print_unclosed_error(self)

    ############################################################################

    async def handle_incoming_message(self, msg: RPCMsg) -> None:
        with self.log_as('handle_incoming_message', msg) as ctx:

            if isinstance(msg, FramedRequestMsg):
                loop = self.event_loop
                send_fn = self.send_msg
                exit_fn = self.on_frame(Incoming, msg)
                task = self.root.exec_incoming_request_task(msg, loop, send_fn, exit_fn)
                self.executing[msg.mid] = task
                # FIXME: we will gradually accumulate finished tasks in executing

            elif isinstance(msg, FramedResponseMsg):
                mid = msg.mid
                awaiting = self.awaiting
                if mid in awaiting:
                    awaiting[mid].set_result(msg)

            # TODO: delete FutureResultMsg in favor of FutureEventMsg
            elif isinstance(msg, FutureResultMsg):
                self.handle_future_msg(msg)

            elif isinstance(msg, FutureEventMsg):
                self.handle_future_event_msg(msg)

            elif isinstance(msg, ChannelMsg):
                await self.channel.put(msg)

            elif isinstance(msg, RemotePrintMsg):
                P.remote_print(msg.text)

            else:
                self.raise_protocol_error(ctx, f'unexpected message: {msg}')

    # TODO: remove in favor of handle_event_msg
    def handle_future_msg(self, msg: FutureResultMsg) -> None:
        with self.log_as('handle_future_msg', msg, msg.fid) as ctx:
            fid = msg.fid
            future = self.futures.pop(fid, None)
            if future is None:
                ctx.warn(f'no future associated with {f_id(fid)}')
                return

            result = msg.data.decode(self.root)
            if result.is_ok:
                ctx.info(f'future.set_result', result.value)
                future.set_result(result.value)
            elif result.is_err:
                ctx.info(f'result.set_error', result.value)
                future.set_exception(result.error)
            elif result.is_unavailable:
                ctx.info(f'future.cancel')
                future.cancel()

    ############################################################################

    async def channel_close(self):
        await self.send_msg(CHANNEL_CLOSED_MSG)

    async def channel_send_value(self, value: Any, *, last: bool = False):
        channel_msg = ChannelMsg.encode(self.root, Result.good(value), last)
        await self.send_msg(channel_msg)

    async def channel_send_exception(self, exception: BaseException, *, last: bool):
        channel_msg = ChannelMsg.encode(self.root, Result.bad(exception), last)
        await self.send_msg(channel_msg)

    async def channel_recv_value(self) -> TermT | Closed:
        result = await self.channel_recv()
        if result is CLOSED:
            return result
        return result.realize()

    async def channel_recv(self) -> Result | Closed:
        channel = self.channel
        if self.closed:
            if channel.empty:
                return CLOSED
        try:
            channel_msg = await channel.get()
        except QueueShutDown:
            return CLOSED
        if channel_msg.last:
            channel.shutdown()  # type: ignore
        return channel_msg.decode(self.root)


REMOTE_PRINT_GUTTER = "\033[43m RP\033[49m "



# fmt: off

from ...core.strs import idsafe_str

from msgspec import msgpack
from re import compile as re_compile
from json import dumps as json_dumps


__all__ = [
    'dec_msgpack',
    'enc_msgpack',
    'fmt_msgpack',
    'pprint_msgpack'
]


################################################################################

dec_msgpack = msgpack.Decoder().decode
enc_msgpack = msgpack.Encoder().encode

################################################################################

def fmt_msgpack(msg: bytes, *, censor_ids: bool = False, multiline: bool = False) -> str:
    try:
        d = dec_msgpack(msg)
    except:
        return '<corrupt msgpack bytes>'
    s = json_dumps(d, indent='\t' if multiline else None, default=_to_json)
    if not multiline:
        s = SL_MID.sub(_fmt_mid, s)
        s = SL_RID.sub(_fmt_rid, s)
    if censor_ids:
        s = idsafe_str(s)
    if multiline:
        s = ML_RID.sub(r'[\1, \2, \3]', s)
    return s

def _to_json(obj):
    if isinstance(obj, bytes):
        return 'bytes:' + repr(obj)
    return '???'

def _fmt_mid(m):
    mid = int(m[1])
    return f'"mid": 0x{mid:x}'

def _fmt_rid(m):
    wid = int(m[1])
    fid = int(m[2])
    rid = int(m[3])
    return f'"rid": [0x{wid:x}, 0x{fid:x}, 0x{rid:x}]]'

################################################################################

def pprint_msgpack(msg: bytes, err: bool = False) -> None:
    from ...core.print import pprint
    data = dec_msgpack(msg)
    pprint(data, err=err)

SL_MID = re_compile(r'"mid": (\d+)')
SL_RID = re_compile(r'"rid": \[(\d+), (\d+), (\d+)]')
ML_RID = re_compile(r'\[\n\s+(\d+),\s+(\d+),\s+([0-9A-Fa-fx]+)\s+]')



# fmt: off

import asyncio
from agentica_internal.core.futures import HookableFuture

from agentica_internal.warpc.worlds.debug_world import *
from agentica_internal.warpc.attrs import FUTURE_ID


async def verify_virtual_future_properties():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    loop = asyncio.get_running_loop()
    async with pair as B:

        future_a = loop.create_future()
        future_b = await B(future_a)

        assert hasattr(future_a, FUTURE_ID)
        assert is_real(future_a)

        assert isinstance(future_b, asyncio.Future)
        assert hasattr(future_b, FUTURE_ID)
        assert is_virtual(future_b)



async def verify_virtual_cancellation():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    loop = asyncio.get_running_loop()
    async with pair as B:

        future_a = loop.create_future()
        future_b = await B(future_a)
        assert is_virtual(future_b)
        assert isinstance(future_b, HookableFuture)

        future_b.cancel()
        await asyncio.sleep(0)
        # ^ because the event msg is merely enqueued to be sent
        assert future_a.cancelled()


async def verify_virtual_completion():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    loop = asyncio.get_running_loop()
    async with pair as B:

        future_a = loop.create_future()
        future_b = await B(future_a)

        future_b.set_result(99)
        await asyncio.sleep(0)
        # ^ because the event msg is merely enqueued to be sent

        assert future_a.done()
        assert future_a.result() == 99


async def verify_real_cancellation():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    loop = asyncio.get_running_loop()
    async with pair as B:

        future_a = loop.create_future()
        future_b = await B(future_a)

        future_a.cancel()
        await asyncio.sleep(0)
        # ^ because the event msg is merely enqueued to be sent

        assert future_b.cancelled()


async def verify_real_completion():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    loop = asyncio.get_running_loop()
    async with pair as B:

        future_a = loop.create_future()
        future_b = await B(future_a)

        future_a.set_result(99)
        await asyncio.sleep(0)
        # ^ because the event msg is merely enqueued to be sent

        assert future_b.done()
        assert future_b.result() == 99


if __name__ == '__main__':
    asyncio.run(verify_virtual_future_properties())
    asyncio.run(verify_virtual_cancellation())
    asyncio.run(verify_virtual_completion())
    asyncio.run(verify_real_cancellation())
    asyncio.run(verify_real_completion())



from . import base


class InvocationError(base.AgenticaError):
    pass


class TooManyInvocationsError(InvocationError):
    pass


class NotRunningError(InvocationError):
    pass


class MalformedInvokeMessageError(InvocationError):
    pass



# fmt: off

import typing
from collections.abc import Iterable, Callable
from typing import Literal, Any, Callable as TyCallable

from .utils import cycle_guard
from ..cpython.classes.sys import *
from ..cpython.classes.anno import *

__all__ = [
    'is_anno',
    'is_anno_class',
    'is_callable_anno',
    'get_callable_return_anno',
    'union_iter',
    'anno_str',
]

####################################################################################################

def is_anno(obj: object):
    cls = type(obj)
    if cls in ANNOS:
        return True
    return cls is type or isinstance(obj, type)

def is_anno_class(cls: type):
    return cls in ANNOS


################################################################################

def is_callable_anno(obj: object) -> bool:
    return obj is Callable or type(obj) in CALLABLES_ANNOS


def get_callable_return_anno(obj: object) -> Any:
    cls = type(obj)
    if cls is TBlankCallable or obj is Callable:
        return Any
    assert cls is TCallable or cls is CCallable, f"{obj} {cls} is not callable"
    args = obj.__args__
    return args[-1]


################################################################################

def union_iter[T](anno: Any, kind: type[T] = object) -> Iterable[T]:
    if isinstance(anno, (TUnion, CUnion)):
        for a in anno.__args__:
            yield from union_iter(a)
    elif isinstance(anno, TAlias):
        yield from union_iter(anno.__value__)
    elif isinstance(anno, kind):
        yield anno

####################################################################################################

type ShowModuleSpec = bool | dict[str, str] | Literal['user']

def anno_str(anno: object, *,
    modules:   bool | dict[str, str] | Literal['user'] = False,
    forward:   Literal['splice', 'quote'] = 'splice',
    typevar:   Literal['name', 'full'] = 'name',
    annotated: Literal['value', 'full'] = 'value',
    aliases:   Literal['name', 'value', 'full'] = 'name',
    simplify:  bool = True,
    private:   bool = False,
    qualify:   bool = False):
    from . import fmt

    """
    Format a type annotation as a string.

    Keyword arguments (the default is the most compact version of all of these):

    * Qualifying modules on classes / annotations like `List[MyClass]`
        * modules='user'                       shows 'List[my_mod.MyClass]'
        * modules=True                         shows `typing.List[my_mod.MyClass]`
        * modules=False                        shows `List[MyClass]`
        * modules={'typing': False, '*': True} shows `List[my_mod.MyClass]`
        * modules={'typing': 'T', '*': False}  shows `T.List[MyClass]`

    * Forward references like `var: 'T'`
        * forward='splice' shows `T`
        * forward='quote'  shows `'T'`
        * forward='full'   shows `ForwardRef['T']`

    * `TypeVar('X', bound=T)` / `def foo[X]`:
        * typevar='short'  shows `Any` if the name is overly complex
        * typevar='name'   shows `X`
        * typevar='full'   shows `~X: T`

    * `Annotated[T, data]`:
        * annotated='value' shows `T`
        * annotated='full'  shows `Annotated[T, data]`
        * annotated='paren' shows `T (data)`

    * `type alias = T`:
        * aliases='name'  shows `alias`
        * aliases='value' shows `T`
        * aliases='full'  shows `alias := T`

    * Vacuous types like `dict[Any, Any]` / `Union[str, None]:
        * simplify=True  shows `dict`           / `Optional[str]
        * simplify=False shows `dict[Any, Any]` / `Union[str, None]`
        
    * Qualified names of classes:
        * qualify=True  shows `func.<locals>.Class` 
        * qualify=False shows `Class`
    """

    _name = _name_fn(modules)

    t_ = _name('typing', '')
    a_ = _name('collections.abc', '')

    if simplify:
        def gen(head: str, args: Iterable[str]):
            args = tuple(args)
            if all(a in _VACUOUS for a in args):
                return head
            body = ', '.join(args)
            return f'{head}[{body}]'
    else:
        def gen(head: str, args: Iterable[str]):
            body = ', '.join(args)
            return f'{head}[{body}]'

    guard = cycle_guard()
    def rec(obj) -> str:

        if obj is Any:
            return f'{t_}Any'

        if obj is NoneT or obj is None:
            return 'None'

        if obj is Ellipsis:
            return '...'

        cls = type(obj)

        if cls is TBlankCallable:
            return f'{t_}Callable'

        if cls is TForm:
            return t_ + obj.__name__

        if cls is type or isinstance(obj, type):
            return _name(obj.__module__, obj.__qualname__ if qualify else obj.__name__)

        if cls is TAlias:
            name = _name(obj.__module__, obj.__name__)
            value = obj.__value__
            cycle = guard(obj)
            if aliases == 'value':
                if cycle:
                    return '...'
                return rec(value)
            if cycle:
                return name
            if aliases == 'name':
                return name
            if aliases == 'full':
                return f'{name} := {rec(value)}'

        if cls is TBlankGeneric:
            return f'{t_}{obj.__name__}'

        if cls is CUnion:
            args = obj.__args__
            return ' | '.join(map(rec, args))

        if cls is TUnpack:
            # TODO: if the arg is a TypedDict, get its fields here
            return rec(obj.__args__[0])

        if cls is TUnion:
            args = obj.__args__
            if simplify and len(args) == 2 and args[1] is NoneT:
                return gen(f'{t_}Optional', map(rec, args[:1]))
            return gen(f'{t_}Union', map(rec, args))

        if cls is TLiteral:
            args = obj.__args__
            body = ', '.join(repr(s) if type(s) is str else rec(s) for s in args)
            return f'{t_}Literal[{body}]'

        if cls in (CGeneric, TGeneric):
            orig = obj.__origin__
            if orig is TyCallable:
                cls = TCallable
            else:
                name = getattr(obj, '_name', None) if cls is TGeneric else None
                name = rec(orig) if name is None else f'{t_}{name}'
                args = obj.__args__
                return gen(name, map(rec, args))

        if cls in (CCallable, TCallable):
            args = obj.__args__
            name = f'{t_}Callable' if cls is TCallable else f'{a_}Callable'
            if not args:
                return name
            *dom, cod = args
            if len(dom) == 1 and type(dom[0]) is EllipT:
                dom = dom[0]
            pair = rec(dom), rec(cod)
            return gen(name, pair)

        if cls is TForward:
            arg = obj.__forward_arg__
            if forward == 'splice':
                return arg
            elif forward == 'quote':
                return repr(arg)
            else:
                return f'{t_}ForwardRef[{arg!r}]'

        if cls is TVar:
            name = obj.__name__
            if typevar == 'name':
                return name
            if typevar == 'short':
                return name if len(name) <= 6 and '_' not in name else 'Any'
            s = '~' + name
            if (bound := obj.__bound__) is not None:
                s += ': ' + rec(bound)
            elif const := obj.__constraints__:
                s += ': ' + rec(const)
            if (default := getattr(obj, '__default__', _NO_DEFAULT)) is not _NO_DEFAULT:
                s += ' = ' + rec(default)
            return s

        if cls is TParamSpec:
            name = _name(obj.__module__, obj.__name__)
            return f'**{name}'

        if cls is TAnnotated:
            arg = rec(obj.__origin__)
            if annotated == 'value':
                return arg
            meta = ', '.join(map(rec, obj.__metadata__))
            return f'{t_}Annotated[{arg}, {meta}]'

        # these are to handle the 'meta' part of Annotated, and list happens in Callable[[A, B], C]
        if cls is list:
            return fmt.f_list(rec, obj)

        if cls is tuple:
            return fmt.f_tuple(rec, obj)

        if cls is dict:
            from .fmt import f_atom
            kv_strs = (f'{f_atom(k)}: {rec(v)}' for k, v in obj.items() if private or not k.startswith('_'))
            return '{' + ', '.join(kv_strs) + '}'

        if cls is set:
            return fmt.f_set(rec, obj)

        if cls in fmt.PRIMITIVES:
            return fmt.f_datum(obj)

        return '...'

    return rec(anno)

def _name_fn(spec: ShowModuleSpec) -> Callable[[str, str], str]:

    match spec:
        case False:
            def fn(_: str, name: str) -> str:
                return name
        case True:
            def fn(mod: str, name: str) -> str:
                if mod and mod not in _BUILTINS:
                    return f'{mod}.{name}'
                return name
        case 'user':
            def fn(mod: str, name: str) -> str:
                if not mod or mod in _NON_USER:
                    return name
                return f'{mod}.{name}'

        case dict():
            get = spec.get
            star = get('*')
            def fn(mod: str, name: str) -> str:
                if mod in _BUILTINS:
                    return name
                mod2 = get(mod, star)
                if mod2 is True:
                    mod2 = mod
                if mod2:
                    return f'{mod2}.{name}'
                return name
        case _:
            raise ValueError(spec)

    return fn


_VACUOUS = '...', 'Any'
_BUILTINS = 'builtins', '__main__'
_NON_USER = {
    'builtins', '__main__', 'typing', 'collections.abc', 'abc', 'enum', 'io', 'pathlib',
    'asyncio'
}
_NO_DEFAULT = getattr(typing, 'NoDefault', None)



# fmt: off

from typing import NamedTuple

from ..shared import *

__all__ = [
    'NamedTupleOf0',
    'NamedTupleOf1',
    'NamedTupleOf2',
    'NamedTupleOf3',
    'NamedTupleWithMeth',
    'NamedTupleViaFunc',
    'NAMED_TUPLE_CLASSES',
    'NAMED_TUPLE_INSTANCES',
]


class NamedTupleOf0(NamedTuple):
    """NamedTupleOf0 docstring"""

    pass


class NamedTupleOf1(NamedTuple):
    """NamedTupleOf1 docstring"""

    a: int


class NamedTupleOf2(NamedTuple):
    """NamedTupleOf2 docstring"""

    a: int
    b: bool


class NamedTupleOf3(NamedTuple):
    """NamedTupleOf2 docstring"""

    a: int
    b: bool
    c: str


class NamedTupleWithMeth(NamedTuple):
    """NamedTupleWithMeth docstring"""

    a: int
    b: bool
    c: str

    def ntup_method(self): ...


NamedTupleViaFunc = NamedTuple(
    'NamedTupleViaFunc',
    [
        ('a', int),
        ('b', bool),
        ('c', str),
    ],
)


NAMED_TUPLE_CLASSES: list[type] = [
    NamedTupleOf0,
    NamedTupleOf1,
    NamedTupleOf2,
    NamedTupleOf3,
    NamedTupleWithMeth,
    NamedTupleViaFunc,
]

NAMED_TUPLE_INSTANCES: list[object] = [
    NamedTupleOf0(),
    NamedTupleOf1(_a),
    NamedTupleOf2(_a, _b),
    NamedTupleOf3(_a, _b, _c),  # type: ignore
    NamedTupleViaFunc(_a, _b, _c),
]



# fmt: off

# NOTE: see core.type for more detail

from types import (
    # singletons
    NoneType                  as NoneT,
    EllipsisType              as EllipT,
    NotImplementedType        as NotImplT,

    # misc
    MappingProxyType          as MapProxyT,
    ModuleType                as ModuleT,
    FrameType                 as FrameT,
    TracebackType             as TracebackT,

    # type annotations
    UnionType                 as UnionT,
    GenericAlias              as GenericT,

    # function types
    CellType                  as CellT,
    CodeType                  as CodeT,
    FunctionType              as FunctionT,
    CoroutineType             as CoroutineT,
    GeneratorType             as GeneratorT,
    AsyncGeneratorType        as AGeneratorT,
    MethodType                as BoundMethodT,

    # wrappers and descriptors
    MethodDescriptorType      as UnboundMethodC,
    WrapperDescriptorType     as UnboundDunderMethodC,
    MethodWrapperType         as BoundDunderMethodC,
    ClassMethodDescriptorType as BoundClassMethodC,
    BuiltinFunctionType       as BoundMethodOrFuncC,
    GetSetDescriptorType      as MutablePropertyC,
    MemberDescriptorType      as SlotPropertyC,
)

__all__ = [
    'NoneT',
    'EllipT',
    'NotImplT',
    'MapProxyT',
    'ModuleT',
    'FrameT',
    'TracebackT',
    'UnionT',
    'GenericT',
    'CellT',
    'CodeT',
    'FunctionT',
    'CoroutineT',
    'GeneratorT',
    'AGeneratorT',
    'BoundMethodT',
    'UnboundMethodC',
    'UnboundDunderMethodC',
    'BoundDunderMethodC',
    'BoundClassMethodC',
    'BoundMethodOrFuncC',
    'MutablePropertyC',
    'SlotPropertyC',

    'PRIMITIVES',
    'CONTAINERS',
    'PY_CALLABLES',
    'C_CALLABLES',
    'CALLABLES',
    'PROPERTIES',

    'SYS_TO_NAME',
    'BUILTIN_TO_NAME'
]

###############################################################################

PY_CALLABLES: tuple[type, ...]
C_CALLABLES:  tuple[type, ...]
CALLABLES:    tuple[type, ...]
PRIMITIVES:   tuple[type, ...]
CONTAINERS:   tuple[type, ...]

PY_CALLABLES = FunctionT, BoundMethodT

C_CALLABLES: tuple[type, ...] = (
    UnboundMethodC,
    UnboundDunderMethodC,
    BoundDunderMethodC,
    BoundClassMethodC,
    BoundMethodOrFuncC,
)

CALLABLES = PY_CALLABLES + C_CALLABLES

PROPERTIES = MutablePropertyC, SlotPropertyC, property

PRIMITIVES = int, bool, float, EllipT, NotImplT, NoneT
CONTAINERS = list, tuple, set, frozenset, dict

SYS_TO_NAME: dict[type, str] = {
    NoneT:                  'NoneT',
    EllipT:                 'EllipT',
    NotImplT:               'NotImplT',
    MapProxyT:              'MapProxyT',
    ModuleT:                'ModuleT',
    FrameT:                 'FrameT',
    TracebackT:             'TracebackT',
    UnionT:                 'UnionT',
    GenericT:               'GenericT',
    CellT:                  'CellT',
    CodeT:                  'CodeT',
    FunctionT:              'FunctionT',
    CoroutineT:             'CoroutineT',
    GeneratorT:             'GeneratorT',
    AGeneratorT:            'AGeneratorT',
    BoundMethodT:           'BoundMethodT',
    UnboundMethodC:         'UnboundMethodC',
    UnboundDunderMethodC:   'UnboundDunderMethodC',
    BoundDunderMethodC:     'BoundDunderMethodC',
    BoundClassMethodC:      'BoundClassMethodC',
    BoundMethodOrFuncC:     'BoundMethodOrFuncC',
    MutablePropertyC:       'MutablePropertyC',
    SlotPropertyC:          'SlotPropertyC',
}

BUILTIN_TO_NAME: dict[type, str] = {
    bool:                   'bool',
    int:                    'int',
    float:                  'float',
    str:                    'str',
    bytearray:              'bytearray',
    bytes:                  'bytes',
    object:                 'object',
    type:                   'type',
    list:                   'list',
    tuple:                  'tuple',
    set:                    'set',
    frozenset:              'frozenset',
    dict:                   'dict'
}



# fmt: off

from .__ import *
from .rpc_request import RequestMsg
from .msg_aliases import ArgsMsg, KwargsMsg

__all__ = [
    'ReplRequestMsg',
    'ReplInitMsg',
    'ReplCallMethodMsg',
    'ReplRunCodeMsg',
]


################################################################################

if TYPE_CHECKING:
    from .term import TermMsg
    from ..request.request_repl import *


################################################################################

class ReplRequestMsg(RequestMsg):

    LOG_TAGS = 'REPL'

    def decode(self, dec: DecoderP) -> 'ReplRequest':
        repl = dec.get_repl()
        try:
            assert repl is not None
            request = self.__decode__(dec)
        except E.WarpDecodingError as error:
            P.eprint("error decoding msg", self)
            P.eprint(error)
            self.pprint(err=True)
            raise
        request.repl = dec.get_repl()
        return request

    def __decode__(self, dec) -> 'ReplRequest':
        raise NotImplementedError()


################################################################################

class ReplInitMsg(ReplRequestMsg, tag='init'):

    global_vars: 'Rec[TermMsg]'
    local_vars: 'Rec[TermMsg]'

    def __decode__(self, dec) -> 'ReplInit':
        from ..request.request_repl import ReplInit
        dec_rec = dec.dec_record
        return ReplInit(dec_rec(self.global_vars), dec_rec(self.local_vars))


################################################################################

class ReplCallMethodMsg(ReplRequestMsg, tag='repl_call_method'):

    method: str
    pos:   'ArgsMsg' = ()
    key:   'KwargsMsg' = {}

    def __decode__(self, dec) -> 'ReplCallMethod':
        from ..request.request_repl import ReplCallMethod
        return ReplCallMethod(self.method, dec.dec_args(self.pos), dec.dec_kwargs(self.key))


################################################################################

class ReplRunCodeMsg(ReplRequestMsg, tag='repl_run_code'):

    source:   str
    options: 'Options'

    def __debug_info_str__(self) -> str:
        return f'{self.source=!r}'

    def __shape__(self) -> str:
        code = self.source
        if len(code) > 32:
            code = code[:32] + '⋯'
        return f'{code!r}'

    def __decode__(self, dec) -> 'ReplRunCode':
        from ..request.request_repl import ReplRunCode
        return ReplRunCode(self.source, **self.options)

    @property
    def is_async(self) -> bool:
        return 'await ' in self.source or 'async ' in self.source



from typing import Any, Literal

from msgspec import UNSET, Struct, UnsetType
from msgspec.json import Decoder, Encoder

from agentica_internal.core.print import pprint

__all__ = [
    #  structures
    "FrameUID",
    "DefnUID",
    "UniMsg",  # data or RPC
    # Data messages
    "ConceptUniMsg",
    "NoDefUniMsg",  # ref or value
    "DefUniMsg",  # resource or annotation
    "DefPayload",
    ## Reference messages
    "RefUniMsg",
    ## Value messages
    "ValueUniMsg",
    "AtomUniMsg",
    "ContainerUniMsg",
    ### Atomic values
    "NoneUniMsg",
    "BoolUniMsg",
    "IntUniMsg",
    "FloatUniMsg",
    "StrUniMsg",
    "BytesUniMsg",
    "TypeStringRefUniMsg",
    ### Symbols
    "EnumValUniMsg",
    ### Container values
    "ArrayUniMsg",
    "SetUniMsg",
    "MapUniMsg",
    "TupleUniMsg",
    ## Resource messages
    "ResourceDefUniMsg",
    "PlaceholderUniMsg",
    "ClassField",
    "ClassMethod",
    "TypeArgument",
    "IndexSignature",
    "ClassPayload",
    "ClassDefUniMsg",
    "FunctionArgument",
    "FunctionPayload",
    "FunctionDefUniMsg",
    "ObjectPayload",
    "ObjectDefUniMsg",
    ## Annotation messages
    "AnnotationDefUniMsg",
    "UnionPayload",
    "UnionUniMsg",
    "IntersectionPayload",
    "IntersectionUniMsg",
    "InterfaceUniMsg",
    "Membership",
    "MethodSignatureUniMsg",
    # RPC messages
    "RpcUniMsg",
    "RequestUniMsg",
    "RequestPayload",
    "ResponseUniMsg",
    "ResponsePayload",
    # Request messages
    "CallArg",
    "CallNewPayload",
    "CallNewUniMsg",
    "CallFunctionPayload",
    "CallFunctionUniMsg",
    "CallMethodPayload",
    "CallMethodUniMsg",
    "HasAttrPayload",
    "HasAttrUniMsg",
    "GetAttrPayload",
    "GetAttrUniMsg",
    "SetAttrPayload",
    "SetAttrUniMsg",
    "DelAttrPayload",
    "DelAttrUniMsg",
    "InstanceOfPayload",
    "InstanceOfUniMsg",
    # Response messages
    "OkUniMsg",
    "ErrPayload",
    "ErrUniMsg",
    "ResPayload",
    "ResUniMsg",
    # Event messages
    "EventUniMsg",
    "FutureCanceledUniMsg",
    "FutureCompletedUniMsg",
    # CODEC
    'json_to_uni',
    'json_to_rpc_uni',
    'json_to_def_uni',
    'json_to_request_uni',
    'json_to_response_uni',
    'uni_to_json',
    'referentialize',
    'make_client_ref',
]

###

type ResourceUniMsg = ClassDefUniMsg | FunctionDefUniMsg | ObjectDefUniMsg
type AnnotationUniMsg = UnionUniMsg | IntersectionUniMsg | InterfaceUniMsg | MethodSignatureUniMsg

# Explicit unions for native decoding
type NoDefMsgUnion = (
    NoneUniMsg
    | BoolUniMsg
    | IntUniMsg
    | FloatUniMsg
    | StrUniMsg
    | TypeStringRefUniMsg
    | BytesUniMsg
    | EnumValUniMsg
    | ArrayUniMsg
    | SetUniMsg
    | MapUniMsg
    | TupleUniMsg
    | RefUniMsg
    | GenericExceptionUniMsg
    | ForeignExceptionUniMsg
    | InternalErrorUniMsg
    | NoDefUniMsg
)

type DefMsgUnion = (
    ClassDefUniMsg
    | FunctionDefUniMsg
    | ObjectDefUniMsg
    | UnionUniMsg
    | IntersectionUniMsg
    | InterfaceUniMsg
    | MethodSignatureUniMsg
    | PlaceholderUniMsg
    | DefUniMsg
)

type RequestUnion = (
    CallNewUniMsg
    | CallFunctionUniMsg
    | CallMethodUniMsg
    | HasAttrUniMsg
    | GetAttrUniMsg
    | SetAttrUniMsg
    | DelAttrUniMsg
    | InstanceOfUniMsg
)

type TermUnion = ValueUniMsg | DefMsgUnion

type ConceptUnion = NoDefMsgUnion | DefMsgUnion

type ResponseUnion = OkUniMsg | ErrUniMsg | ResUniMsg

type EventUnion = FutureCanceledUniMsg | FutureCompletedUniMsg

type RpcUnion = RequestUnion | ResponseUnion | EventUnion

type AnyUni = NoDefMsgUnion | DefMsgUnion | RpcUnion

type FrameID = int
type DefnID = int

type WorldEnum = Literal['client', 'server']
type MembershipKind = Literal['static', 'instance']

###


class FrameUID(Struct):
    world: WorldEnum
    frame: FrameID


class DefnUID(Struct):
    world: WorldEnum
    resource: DefnID
    # Added merely for Python compatibility
    py_world: int = -1
    py_resource: int = -1
    py_frame: int = -1

    def __hash__(self) -> int:  # make hashable for dict/set usage
        return hash((self.world, self.resource, self.py_world, self.py_resource, self.py_frame))


###


class UniMsg(Struct, tag_field='kind'):  # data or RPC
    def __short_str__(self) -> str:
        """Determines how this will get formatted in stack traces."""
        return type(self).__name__ + '(...)'

    def pprint(self) -> None:
        """Used in logging."""
        pprint(json_to_python(uni_to_json(self)))


class ConceptUniMsg(UniMsg):
    pass


class NoDefUniMsg(ConceptUniMsg):  # ref or value
    pass


class DefPayload(Struct, kw_only=True):
    is_top_level: bool = False


class DefUniMsg(ConceptUniMsg):  # resource or annotation
    uid: DefnUID
    payload: DefPayload


# Take a concept, churn out a ref
def referentialize(msg: ConceptUniMsg) -> NoDefUniMsg:
    # HACK: Pass through atoms (StrUniMsg, IntUniMsg, etc.) unchanged
    # This allows TypeScript to embed literal values in TypeArgument.type for typing.Literal support
    if isinstance(msg, NoDefUniMsg):
        return msg
    elif isinstance(msg, ClassDefUniMsg):
        return RefUniMsg(uid=msg.uid, system=msg.payload.system_resource)
    elif isinstance(msg, DefUniMsg):
        return RefUniMsg(uid=msg.uid)
    else:
        raise ValueError(f'Unknown DataUniMsg: {msg}')


def make_client_ref(uid: int) -> 'RefUniMsg':
    return RefUniMsg(uid=DefnUID(world='client', resource=uid))


###


class RefUniMsg(NoDefUniMsg, tag='ref', tag_field='kind'):
    uid: DefnUID
    system: bool = False


###


class ValueUniMsg(NoDefUniMsg):
    pass


###


class AtomUniMsg(ValueUniMsg):
    pass


class NoneUniMsg(AtomUniMsg, tag='none', tag_field='kind'):
    val: None = None


class BoolUniMsg(AtomUniMsg, tag='bool', tag_field='kind'):
    val: bool


class IntUniMsg(AtomUniMsg, tag='int', tag_field='kind'):
    val: int


class FloatUniMsg(AtomUniMsg, tag='float', tag_field='kind'):
    val: float


class StrUniMsg(AtomUniMsg, tag='str', tag_field='kind'):
    val: str


class BytesUniMsg(AtomUniMsg, tag='bytes', tag_field='kind'):
    val: bytes


###


class TypeStringRefUniMsg(RefUniMsg, tag='type_string', tag_field='kind'):
    uid: DefnUID  # Matches RefUniMsg protocol
    name: str = ''


###


class EnumValUniMsg(ValueUniMsg, tag='enum_val', tag_field='kind'):
    val: StrUniMsg | IntUniMsg | FloatUniMsg
    cls: RefUniMsg


###


class ContainerUniMsg(ValueUniMsg):
    pass


class ArrayUniMsg(ContainerUniMsg, tag='array', tag_field='kind'):
    val: list[NoDefMsgUnion]


class SetUniMsg(ContainerUniMsg, tag='set', tag_field='kind'):
    val: list[NoDefMsgUnion]


class MapPayload(Struct):
    keys: list[NoDefMsgUnion]
    vals: list[NoDefMsgUnion]


class MapUniMsg(ContainerUniMsg, tag='map', tag_field='kind'):
    val: MapPayload


class TupleUniMsg(ContainerUniMsg, tag='tuple', tag_field='kind'):
    val: list[NoDefMsgUnion]


###


class GenericExceptionUniMsg(NoDefUniMsg, tag='generic_excp', tag_field='kind', kw_only=True):
    excp_cls_name: str
    excp_str_args: list[str]
    excp_stack: list[str]


class ForeignExceptionUniMsg(NoDefUniMsg, tag='foreign_excp', tag_field='kind', kw_only=True):
    excp_cls: RefUniMsg
    excp_args: list[NoDefMsgUnion]


class InternalErrorUniMsg(NoDefUniMsg, tag='internal_error', tag_field='kind', kw_only=True):
    error: str


###


class ResourceDefUniMsg(DefUniMsg):
    uid: DefnUID
    payload: Any


###


class PlaceholderUniMsg(ResourceDefUniMsg):
    payload: dict[str, Any] | UnsetType = UNSET


###


class FunctionArgument(Struct):
    name: str
    type: RefUniMsg
    optional: bool | UnsetType = UNSET
    default: NoDefMsgUnion | UnsetType = UNSET
    rest: bool | UnsetType = UNSET


class FunctionPayload(DefPayload):
    name: str
    arguments: list[FunctionArgument]
    doc: str | UnsetType = UNSET
    returnType: RefUniMsg | UnsetType = UNSET
    module: str | UnsetType = UNSET


class FunctionDefUniMsg(ResourceDefUniMsg, tag='func', tag_field='kind'):
    uid: DefnUID
    payload: FunctionPayload


###


class ClassField(Struct):
    name: str
    type: RefUniMsg
    is_private: bool | UnsetType = UNSET
    is_static: bool | UnsetType = UNSET
    default: NoDefMsgUnion | UnsetType = UNSET
    is_optional: bool | UnsetType = UNSET


class ClassMethod(Struct):
    name: str
    function: RefUniMsg
    is_private: bool | UnsetType = UNSET
    is_static: bool | UnsetType = UNSET


class TypeArgument(Struct):
    name: str
    # HACK: May be AtomUniMsgs for Literal types
    type: RefUniMsg | StrUniMsg | IntUniMsg | FloatUniMsg | BoolUniMsg


class IndexSignature(Struct):
    key_type: RefUniMsg
    value_type: RefUniMsg
    map_type: RefUniMsg


class ClassPayload(DefPayload):
    name: str
    fields: list[ClassField] | UnsetType = UNSET
    methods: list[ClassMethod] | UnsetType = UNSET
    doc: str | UnsetType = UNSET
    module: str | UnsetType = UNSET
    ctor_args: list["FunctionArgument"] | UnsetType = UNSET
    supplied_type_args: list["TypeArgument"] | FunctionDefUniMsg | UnsetType = UNSET
    instance_of_generic: DefnUID | UnsetType = UNSET
    bases: list[RefUniMsg] | UnsetType = UNSET
    system_resource: bool = False
    is_plain_object_type: bool | UnsetType = UNSET
    index_signature: IndexSignature | UnsetType = UNSET


class ClassDefUniMsg(ResourceDefUniMsg, tag='cls', tag_field='kind'):
    uid: DefnUID
    payload: ClassPayload


###


class ObjectPayload(DefPayload):
    name: str | UnsetType = UNSET
    cls: RefUniMsg | UnsetType = UNSET  # NOTE: note always a class, could be an annotation
    keys: list[str] | UnsetType = UNSET


class ObjectDefUniMsg(ResourceDefUniMsg, tag='obj', tag_field='kind'):
    uid: DefnUID
    payload: ObjectPayload


###


# Annotation messages
class AnnotationDefUniMsg(DefUniMsg):
    uid: DefnUID
    payload: Any


###


class UnionPayload(DefPayload):
    classes: list[RefUniMsg]


class UnionUniMsg(AnnotationDefUniMsg, tag='union', tag_field='kind'):
    uid: DefnUID
    payload: UnionPayload


###


class IntersectionPayload(DefPayload):
    classes: list[RefUniMsg]


class IntersectionUniMsg(AnnotationDefUniMsg, tag='intersection', tag_field='kind'):
    uid: DefnUID
    payload: IntersectionPayload


###


class InterfaceUniMsg(AnnotationDefUniMsg, tag='interface', tag_field='kind'):
    uid: DefnUID
    payload: ClassPayload


###


class Membership(Struct):
    uid: DefnUID
    kind: MembershipKind


class MethodSignatureUniMsg(AnnotationDefUniMsg, tag='methodsig', tag_field='kind'):
    uid: DefnUID
    methodOf: Membership
    payload: FunctionPayload


def toMethodSignature(msg: FunctionDefUniMsg, mem: Membership) -> MethodSignatureUniMsg:
    return MethodSignatureUniMsg(uid=msg.uid, methodOf=mem, payload=msg.payload)


###


# RPC messages
class RpcUniMsg(UniMsg, tag_field='kind', kw_only=True):
    pass


class RequestPayload(Struct):
    pass


class RequestUniMsg(RpcUniMsg, tag_field='kind', kw_only=True):
    payload: RequestPayload
    parentFID: FrameID = -1
    selfFID: FrameID = -1
    requestedFID: FrameID = -1
    defs: list[DefMsgUnion] = []


class ResponsePayload(Struct):
    pass


class ResponseUniMsg(RpcUniMsg, tag_field='kind', kw_only=True):
    parentFID: FrameID = -1
    selfFID: FrameID = -1
    iid: str | UnsetType = UNSET


###


class CallArg(Struct):
    name: str  # For bookkeeping
    val: NoDefMsgUnion


class CallNewPayload(RequestPayload):
    cls: RefUniMsg
    args: list[CallArg]
    type_args: dict[str, RefUniMsg] | UnsetType = UNSET


class CallNewUniMsg(RequestUniMsg, tag='new', tag_field='kind'):
    payload: CallNewPayload


###


class CallFunctionPayload(RequestPayload):
    fun: RefUniMsg
    args: list[CallArg]


class CallFunctionUniMsg(RequestUniMsg, tag='callfun', tag_field='kind'):
    payload: CallFunctionPayload


###


class CallMethodPayload(RequestPayload):
    owner: RefUniMsg
    method_name: str
    args: list[CallArg]
    method_ref: RefUniMsg | UnsetType = UNSET


class CallMethodUniMsg(RequestUniMsg, tag='callmethod', tag_field='kind'):
    payload: CallMethodPayload


###


class HasAttrPayload(RequestPayload):
    owner: RefUniMsg
    attr: str | None


class HasAttrUniMsg(RequestUniMsg, tag='hasattr', tag_field='kind'):
    payload: HasAttrPayload


###


class GetAttrPayload(RequestPayload):
    owner: RefUniMsg
    attr: str | None


class GetAttrUniMsg(RequestUniMsg, tag='getattr', tag_field='kind'):
    payload: GetAttrPayload


###


class SetAttrPayload(RequestPayload):
    owner: RefUniMsg
    attr: str | None
    val: NoDefMsgUnion


class SetAttrUniMsg(RequestUniMsg, tag='setattr', tag_field='kind'):
    payload: SetAttrPayload


###


class DelAttrPayload(RequestPayload):
    owner: RefUniMsg
    attr: str | None


class DelAttrUniMsg(RequestUniMsg, tag='delattr', tag_field='kind'):
    payload: DelAttrPayload


###


class InstanceOfPayload(RequestPayload):
    owner: RefUniMsg
    cls: RefUniMsg


class InstanceOfUniMsg(RequestUniMsg, tag='instanceof', tag_field='kind'):
    payload: InstanceOfPayload


###


# Response messages
class OkUniMsg(ResponseUniMsg, tag='ok', tag_field='kind', kw_only=True):
    payload: None = None


###


class ErrPayload(ResponsePayload):
    error: ForeignExceptionUniMsg | InternalErrorUniMsg | GenericExceptionUniMsg


class ErrUniMsg(ResponseUniMsg, tag='err', tag_field='kind', kw_only=True):
    payload: ErrPayload


###


class ResPayload(ResponsePayload):
    result: DefMsgUnion | NoDefMsgUnion


class ResUniMsg(ResponseUniMsg, tag='result', tag_field='kind', kw_only=True):
    payload: ResPayload
    defs: list[DefMsgUnion] = []


###


class EventUniMsg(RpcUniMsg, tag_field='kind'):
    future_id: int


class FutureCanceledUniMsg(EventUniMsg, tag='future_canceled', tag_field='kind', kw_only=True):
    future_id: int


class FutureCompletedUniMsg(EventUniMsg, tag='future_completed', tag_field='kind'):
    future_id: int
    result: ResUniMsg | ErrUniMsg


json_to_uni = Decoder(AnyUni).decode
json_to_rpc_uni = Decoder(RpcUnion).decode
json_to_concept_uni = Decoder(ConceptUnion).decode
json_to_def_uni = Decoder(DefMsgUnion).decode
json_to_term_uni = Decoder(TermUnion).decode
json_to_request_uni = Decoder(RequestUnion).decode
json_to_response_uni = Decoder(ResponseUnion).decode
uni_to_json = Encoder().encode
json_to_python = Decoder().decode



# fmt: off

from collections.abc import Callable
from functools import partial

__all__ = [
    'log0',
    'log1',
    'log2',
    'log_fn',
    'no_log_fn'
]

###############################################################################

def log0(*args) -> None:
    from ..print import tprint
    tprint(*args)

def log1(label: str, *args) -> None:
    from ..print import colorize, tprint
    tprint(colorize(label), *args)

def log2(label: str, label2: str, *args) -> None:
    from ..print import colorize, tprint
    tprint(colorize(label), colorize(label2), *args)

def log_fn(label: str | tuple[str, ...]) -> Callable[..., None]:
    from ..print import colorize, tprint
    if isinstance(label, str):
        labels = colorize(label),
    else:
        labels = map(colorize, label)

    return partial(tprint, *labels)

def no_log_fn(*_) -> None:
    pass



# fmt: off

from types import ModuleType

from . import _empty as empty_module
from . import _simple as simple_module

__all__ = [
    'simple_module',
    'empty_module',
    'manual_module',
    'impostor_module_1',
    'impostor_module_2',
    'holdout_module',
    'MODULES',
]


manual_module = ModuleType('manual_module')
setattr(manual_module, 'i', 5)
setattr(manual_module, 'b', True)
setattr(manual_module, '__annotations__', {'x': int})

# these are designed to be *almost* the same
impostor_module_1 = ModuleType('impostor_module')
setattr(impostor_module_1, 'i', 6)
setattr(impostor_module_1, 'b', True)
setattr(impostor_module_1, '__annotations__', {'x': int})

impostor_module_2 = ModuleType('impostor_module')
setattr(impostor_module_2, 'i', 7)
setattr(impostor_module_2, 'b', True)
setattr(impostor_module_2, '__annotations__', {'x': int})

# this isn't in `MODULES`, so that it can be used as __ne__ comparison
holdout_module = ModuleType('holdout_module')
setattr(holdout_module, 's', 'hello')
setattr(holdout_module, 'b', False)
setattr(holdout_module, 'l', [1, 2, 3])
setattr(holdout_module, '__annotations__', {'s': str, 'b': bool, 'l': list})


MODULES = [
    simple_module,
    empty_module,
    manual_module,
    impostor_module_1,
    impostor_module_2,
]



import builtins
from asyncio import Future
from collections import defaultdict
from enum import Enum
from types import FunctionType
from typing import TYPE_CHECKING, Any, Literal, Never
from typing import ClassVar as TypingClassVar

from msgspec import UNSET, UnsetType

from agentica_internal.core.print import colorize, hprint
from agentica_internal.core.collections import bidict
from agentica_internal.core.log.log_flag import LogFlag
from agentica_internal.warpc.alias import GlobalRID
from agentica_internal.warpc.system import LRID_TO_SRID

from .uni_msgs import (
    ClassDefUniMsg,
    ClassPayload,
    DefnUID,
    DefUniMsg,
    InterfaceUniMsg,
    IntersectionPayload,
    IntersectionUniMsg,
    RefUniMsg,
    TypeStringRefUniMsg,
)
from .uni_sys_id import BUILTIN_UNI_IDS

if TYPE_CHECKING:
    from .uni_msgs import DefnUID, FunctionDefUniMsg, Membership, MethodSignatureUniMsg

CNV = 1048576  # conversion factor for resource IDs (=2^20)
ORDER_MARKER = "|"

LOG_TS = LogFlag('TS')

DEP_GRAPH = colorize('DepGraph')


def get_def_or_sysref_from_ctx(
    ctx: dict[DefnUID, DefUniMsg], ref: 'RefUniMsg | Membership | TypeStringRefUniMsg'
) -> DefUniMsg | RefUniMsg | None:
    from .uni_msgs import Membership, RefUniMsg, TypeStringRefUniMsg

    if isinstance(ref, (RefUniMsg, Membership)) and not isinstance(ref, TypeStringRefUniMsg):
        uid = ref.uid
    elif isinstance(ref, TypeStringRefUniMsg):
        uid = ref.uid
    else:
        raise ValueError(f"Transcoder Error: Unsupported reference type: {type(ref)}")

    deref = ctx.get(uid, None)

    if deref is None:
        if uid.resource < 0:
            # For system resources use references in place of defs
            return RefUniMsg(uid=uid)
        else:
            raise ValueError(f"Transcoder Error: Reference not found in context: {uid}")
    return deref


def gridToResourceUID(grid: GlobalRID) -> 'DefnUID':
    # TODO: @tali ... hack: use -1 has a marker for client world

    from .uni_msgs import DefnUID

    if grid[0] == -1:
        world = 'client'
        resource_UID = grid[2]
        return DefnUID(
            world=world,
            resource=resource_UID,
        )
    else:
        world = 'server'
        # Try to generate somewhat unique deterministic resource ID
        resource_UID = (
            int(grid[0] % CNV) * CNV * CNV + int(grid[1] % CNV) * CNV + (grid[2] % CNV) - CNV
        )
        return DefnUID(
            world=world,
            resource=resource_UID,
            py_world=grid[0],
            py_frame=grid[1],
            py_resource=grid[2],
        )


def resourceUIDToGrid(uid: 'DefnUID') -> GlobalRID:
    # TODO: @tali here's the reverse hack
    if uid.world == 'client':
        return (-1, -1, uid.resource)
    else:
        assert uid.world == 'server'
        assert uid.py_world is not UNSET
        assert uid.py_resource is not UNSET
        assert uid.py_frame is not UNSET
        return (uid.py_world, uid.py_resource, uid.py_frame)


def unsetNone[T](value: T | None) -> T | UnsetType:
    if value is None:
        return UNSET
    return value


def function_to_method(
    function: 'FunctionDefUniMsg', methodOf: 'Membership'
) -> 'MethodSignatureUniMsg':
    from .uni_msgs import MethodSignatureUniMsg

    return MethodSignatureUniMsg(
        uid=function.uid,
        methodOf=methodOf,
        payload=function.payload,
    )


###
# System resource conversion
# TODO: this is neither efficient nor complete atm

_PY_SYSID_TO_UNI_NAME: dict[int, set[str]] = {
    id(object): {'Object'},
    id(str): {'String'},
    id(int): {'Number'},
    id(float): {'Number'},
    id(bool): {'Boolean'},
    id(list): {'Array'},
    id(dict): {'Map'},
    id(set): {'Set'},
    id(tuple): {'Tuple'},
    id(Exception): {'Error'},
    id(BaseException): {'Error'},
    id(Future): {'Future'},
    id(FunctionType): {'Function'},
    id(type(None)): {'None'},
    id(Never): {'Never'},
    id(Literal): {'Literal'},
    id(AttributeError): {'AttributeError'},
    id(KeyError): {'KeyError'},
    id(ValueError): {'ValueError'},
    id(TypeError): {'TypeError'},
    id(builtins.repr): {'magicRepr'},
    id(builtins.print): {'console.log'},
    id(builtins.len): {'magicLen'},
    id(builtins.isinstance): {'magicInstanceof'},
    id(builtins.hasattr): {'Reflect.has'},
    id(builtins.getattr): {'Reflect.get'},
    id(builtins.dir): {'magicKeys'},
    id(Enum): {'Enum'},
    id(type): {'TYPE'},
    id(TypingClassVar): {'ClassVar'},
}


def sridToUniSysId(sid: int) -> int | None:
    for py_sys_id, uni_names in _PY_SYSID_TO_UNI_NAME.items():
        if LRID_TO_SRID.get(py_sys_id) == sid:
            for uni_name in uni_names:
                uni_id = BUILTIN_UNI_IDS.get(uni_name)
                if uni_id is not None:
                    return uni_id
    return None


def uniSysIdToSrid(uni_id: int) -> int | None:
    uni_name: str | None = None
    for name, uid in BUILTIN_UNI_IDS.items():
        if uid == uni_id:
            uni_name = name
            break
    if not uni_name:
        return None
    for py_sys_id, mapped_names in _PY_SYSID_TO_UNI_NAME.items():
        if uni_name in mapped_names:
            return LRID_TO_SRID.get(py_sys_id)
    return None


class MangleManager:
    """
    Original names are deterministically mapped to normalized names.
    This mapping is not injective, so we need to add a suffix to the normalized name
    to obtain unique names.
    """

    type NormalizedName = str
    type Suffix = int
    type OriginalName = str
    type UniqueName = str

    uniqueness: dict[NormalizedName, Suffix]
    names: bidict[OriginalName, UniqueName]

    # special-case normalization rules
    normalizers: list[tuple[str, str]]
    reserved_prefixes: list[str]
    reserved_suffixes: list[str]

    NUM_PREFIX = 'NUM__'
    KW_SUFFIX = '_KW_'

    def __init__(self):
        self.uniqueness = defaultdict(int)
        self.names = bidict()
        # default special-case TypeScript->Python normalization rules
        self.normalizers = [
            ('$', 'Dollar_'),
        ]
        self.reserved_prefixes = [self.NUM_PREFIX]
        self.reserved_suffixes = [self.KW_SUFFIX]

    def normalize(self, original_name: OriginalName) -> NormalizedName:
        """replace non-Python compatible characters with underscores, and guard for keywords"""
        import keyword
        import re

        # check the ident does not already use a reserved prefix or suffix or sentinel
        uses_reserved_prefix = any(map(original_name.startswith, self.reserved_prefixes))
        uses_reserved_suffix = any(map(original_name.endswith, self.reserved_suffixes))
        uses_sentinel = any(sentinel in original_name for _, sentinel in self.normalizers)
        if uses_reserved_prefix or uses_reserved_suffix or uses_sentinel:
            raise ValueError(
                f"Identifier {original_name} cannot use reserved prefixes"
                + f"({', '.join(self.reserved_prefixes)})"
                + f", suffixes "
                + f"({', '.join(self.reserved_suffixes)})"
                + f", or sentinels "
                + f"({', '.join(sentinel for _, sentinel in self.normalizers)})"
            )

        normalized_name = original_name

        # we special-case a few normalization rules first.
        for original_pattern, normalized_pattern in self.normalizers:
            normalized_name = normalized_name.replace(original_pattern, normalized_pattern)

        # remaining invalid characters are replaced with underscores
        normalized = re.sub(r"\W", "_", normalized_name)

        if not normalized[0].isalpha() and normalized[0] != '_':
            normalized = f"{self.NUM_PREFIX}{normalized}"

        if keyword.iskeyword(normalized):
            normalized = f"{normalized}{self.KW_SUFFIX}"

        return normalized

    def mangle(self, original_name: OriginalName) -> UniqueName:
        if original_name in self.names:
            # this name already has a unique mangled name
            return self.names[original_name]

        normalized_name = self.normalize(original_name)
        self.names[original_name] = normalized_name
        return normalized_name

    def demangle(self, unique_name: UniqueName) -> OriginalName:
        """reverse the mangling process if this name was previously mangled."""
        # leave unchanged if never mangled
        return self.names.with_value(unique_name, default=unique_name)


def msg_uni_name_sanitization(msg: Any, mangler: MangleManager) -> Any:
    """
    Recursively sanitize all names in a UniMsg by mangling TypeScript names into
    Python-compatible identifiers using the provided MangleManager.
    Used when receiving messages from TypeScript.
    """
    from .uni_msgs import (
        CallMethodPayload,
        ClassField,
        ClassMethod,
        ClassPayload,
        DelAttrPayload,
        FunctionArgument,
        FunctionPayload,
        GetAttrPayload,
        HasAttrPayload,
        SetAttrPayload,
    )

    if msg is UNSET or msg is None:
        return msg

    # Some fields are lists
    if isinstance(msg, list):
        return [msg_uni_name_sanitization(item, mangler) for item in msg]

    # Handle structs with __struct_fields__
    fields = getattr(type(msg), "__struct_fields__", None)
    if not fields:
        return msg

    updates = {}

    # Special handling for specific payload types with name fields
    if isinstance(msg, ClassPayload):
        if msg.name is not UNSET:
            updates['name'] = mangler.mangle(msg.name)
    elif isinstance(msg, FunctionPayload):
        if msg.name is not UNSET:
            updates['name'] = mangler.mangle(msg.name)
    elif isinstance(msg, ClassMethod):
        if msg.name is not UNSET:
            updates['name'] = mangler.mangle(msg.name)
    elif isinstance(msg, ClassField):
        if msg.name is not UNSET:
            updates['name'] = mangler.mangle(msg.name)
    elif isinstance(msg, FunctionArgument):
        if msg.name is not UNSET:
            updates['name'] = mangler.mangle(msg.name)
    elif isinstance(msg, (GetAttrPayload, SetAttrPayload, DelAttrPayload, HasAttrPayload)):
        if hasattr(msg, 'attr') and msg.attr is not None:
            updates['attr'] = mangler.mangle(msg.attr)
    elif isinstance(msg, CallMethodPayload):
        if msg.method_name is not UNSET:
            updates['method_name'] = mangler.mangle(msg.method_name)

    # Recursively process all fields
    for field_name in fields:
        if field_name in updates:
            continue  # Already handled above
        try:
            attr = getattr(msg, field_name)
        except Exception:
            continue
        if attr is not UNSET:
            new_attr = msg_uni_name_sanitization(attr, mangler)
            if new_attr is not attr:
                updates[field_name] = new_attr

    if updates:
        return type(msg)(**{**{f: getattr(msg, f) for f in fields}, **updates})
    return msg


def msg_TS_name_restoration(msg: Any, mangler: MangleManager) -> Any:
    """
    Recursively restore all names in a UniMsg by demangling Python identifiers
    back to their original TypeScript names using the provided MangleManager.
    Used when sending messages to TypeScript.
    """
    from .uni_msgs import (
        CallMethodPayload,
        ClassField,
        ClassMethod,
        ClassPayload,
        DelAttrPayload,
        FunctionArgument,
        FunctionPayload,
        GetAttrPayload,
        HasAttrPayload,
        SetAttrPayload,
    )

    if msg is UNSET or msg is None:
        return msg

    # Some fields are lists
    if isinstance(msg, list):
        return [msg_TS_name_restoration(item, mangler) for item in msg]

    # Handle structs with __struct_fields__
    fields = getattr(type(msg), "__struct_fields__", None)
    if not fields:
        return msg

    updates = {}

    # Special handling for specific payload types with name fields
    if isinstance(msg, ClassPayload):
        if msg.name is not UNSET:
            updates['name'] = mangler.demangle(msg.name)
    elif isinstance(msg, FunctionPayload):
        if msg.name is not UNSET:
            updates['name'] = mangler.demangle(msg.name)
    elif isinstance(msg, ClassMethod):
        if msg.name is not UNSET:
            updates['name'] = mangler.demangle(msg.name)
    elif isinstance(msg, ClassField):
        if msg.name is not UNSET:
            updates['name'] = mangler.demangle(msg.name)
    elif isinstance(msg, FunctionArgument):
        if msg.name is not UNSET:
            updates['name'] = mangler.demangle(msg.name)
    elif isinstance(msg, (GetAttrPayload, SetAttrPayload, DelAttrPayload, HasAttrPayload)):
        if hasattr(msg, 'attr') and msg.attr is not None:
            updates['attr'] = mangler.demangle(msg.attr)
    elif isinstance(msg, CallMethodPayload):
        if msg.method_name is not UNSET:
            updates['method_name'] = mangler.demangle(msg.method_name)

    # Recursively process all fields
    for field_name in fields:
        if field_name in updates:
            continue  # Already handled above
        try:
            attr = getattr(msg, field_name)
        except Exception:
            continue
        if attr is not UNSET:
            new_attr = msg_TS_name_restoration(attr, mangler)
            if new_attr is not attr:
                updates[field_name] = new_attr

    if updates:
        return type(msg)(**{**{f: getattr(msg, f) for f in fields}, **updates})
    return msg


def order_defs(defs: list[DefUniMsg], keys: list[Any] = []) -> tuple[list[DefUniMsg], list[Any]]:
    """
    Topologically sort definitions by dependencies.

    This algorithm handles cycles by:
    1. Detecting inheritance cycles (which are errors)
    2. Grouping mutually recursive class-like defs into strata
    3. Sorting strata by dependency order
    4. Within each stratum, processing defs in inheritance order
    5. Replacing forward references with TypeStringRefUniMsg where needed

    Bases references are never replaced with type strings (they must be concrete types).
    """

    uid_to_def: dict[DefnUID, DefUniMsg] = {definition.uid: definition for definition in defs}

    def is_classlike(defmsg: DefUniMsg) -> bool:
        """Class-like defs (ClassDefUniMsg, InterfaceUniMsg, IntersectionUniMsg) can be replaced by type strings"""
        return (
            isinstance(defmsg, (ClassDefUniMsg, InterfaceUniMsg, IntersectionUniMsg))
            and defmsg.uid.resource > -1
        )

    def get_name(defmsg: DefUniMsg) -> str:
        """Extract the name from a definition."""
        return getattr(defmsg.payload, 'name', f'Unnamed{defmsg.uid.resource}')

    def walk_refs(value: Any, callback: Any) -> None:
        """
        Recursively walk a structure and invoke callback on each RefUniMsg found.
        Does not walk into TypeStringRefUniMsg.
        """
        if value is UNSET or value is None:
            return
        if isinstance(value, RefUniMsg):
            callback(value)
            return
        if isinstance(value, TypeStringRefUniMsg):
            return
        if isinstance(value, (list, tuple, set)):
            for item in value:
                walk_refs(item, callback)
            return
        if isinstance(value, dict):
            for k, v in value.items():
                walk_refs(k, callback)
                walk_refs(v, callback)
            return
        fields = getattr(type(value), "__struct_fields__", None)
        if fields:
            for field_name in fields:
                try:
                    attr = getattr(value, field_name)
                except Exception:
                    continue
                if attr is not UNSET:
                    walk_refs(attr, callback)

    def extract_refs(defmsg: DefUniMsg) -> set[DefnUID]:
        """
        Extract all RefUniMsg UIDs from a definition.
        """
        found: set[DefnUID] = set()
        allowed_uids = set(uid_to_def.keys())

        def process_ref(ref: RefUniMsg) -> None:
            if ref.uid in allowed_uids:
                found.add(ref.uid)

        walk_refs(defmsg, process_ref)

        return found

    def extract_bases(defmsg: DefUniMsg) -> set[DefnUID]:
        """Extract classlike UIDs from the 'bases' or 'classes' field of a class-like definition."""
        if not is_classlike(defmsg):
            return set()

        if isinstance(defmsg, IntersectionUniMsg):
            bases_list = getattr(defmsg.payload, 'classes', UNSET)
        else:
            bases_list = getattr(defmsg.payload, 'bases', UNSET)

        if bases_list is UNSET or bases_list is None:
            return set()

        found: set[DefnUID] = set()

        def process_ref(ref: RefUniMsg) -> None:
            if ref.uid in classlike_uids:
                found.add(ref.uid)

        walk_refs(bases_list, process_ref)
        return found

    def replace_refs_in_def(defmsg: DefUniMsg, uids_to_replace: dict[DefnUID, str]) -> DefUniMsg:
        """
        Create a copy of defmsg with specified UIDs replaced by TypeStringRefUniMsg.
        Should never replaces refs in the 'bases' field.
        """
        if not uids_to_replace:
            return defmsg

        def replacer(value: Any) -> tuple[Any, bool]:
            """Returns (new_value, changed)."""
            if value is UNSET or value is None:
                return value, False
            if isinstance(value, RefUniMsg):
                if value.uid in uids_to_replace:
                    return TypeStringRefUniMsg(uid=value.uid, name=uids_to_replace[value.uid]), True
                return value, False
            if isinstance(value, (list, tuple, set)):
                result = []
                any_changed = False
                for item in value:
                    new_item, changed = replacer(item)
                    result.append(new_item)
                    any_changed = any_changed or changed
                if any_changed:
                    return (type(value)(result) if not isinstance(value, list) else result), True
                return value, False
            if isinstance(value, dict):
                new_dict = {}
                any_changed = False
                for k, v in value.items():
                    new_v, changed = replacer(v)
                    new_dict[k] = new_v
                    any_changed = any_changed or changed
                if any_changed:
                    return new_dict, True
                return value, False
            fields = getattr(type(value), "__struct_fields__", None)
            if fields:
                updates = {}
                for field_name in fields:
                    try:
                        attr = getattr(value, field_name)
                    except Exception:
                        continue
                    if attr is not UNSET:
                        new_attr, changed = replacer(attr)
                        if changed:
                            updates[field_name] = new_attr
                if updates:
                    if 'bases' in updates and isinstance(value, ClassPayload):
                        raise RuntimeError(
                            f"Forward reference detection attempted to replace 'bases' of class {get_name(defmsg)} with type string."
                        )
                    if 'classes' in updates and isinstance(value, IntersectionPayload):
                        raise RuntimeError(
                            f"Forward reference detection attempted to replace 'bases classes' of interface {get_name(defmsg)} with type string."
                        )
                    return type(value)(
                        **{**{f: getattr(value, f) for f in fields}, **updates}
                    ), True
            return value, False

        result, _ = replacer(defmsg)
        return result

    # ===== STAGE 0: Build dependency and inheritance graphs =====

    dependency_graph: dict[DefnUID, set[DefnUID]] = {}
    inheritance_graph: dict[DefnUID, set[DefnUID]] = {}
    classlike_uids: set[DefnUID] = set()

    for defmsg in defs:
        uid = defmsg.uid
        dependency_graph[uid] = extract_refs(defmsg)

        if is_classlike(defmsg):
            classlike_uids.add(uid)
            inheritance_graph[uid] = extract_bases(defmsg)

    if LOG_TS:
        hprint(DEP_GRAPH, f"Built graphs: {len(defs)} defs, {len(classlike_uids)} classlike")

    # ===== STAGE 1: Detect inheritance cycles =====

    visited_inheritance: set[DefnUID] = set()

    def find_inheritance_cycle(start_uid: DefnUID) -> list[DefnUID] | None:
        """DFS to find a cycle in the inheritance graph starting from start_uid."""
        path: list[DefnUID] = []

        def dfs(current_uid: DefnUID) -> list[DefnUID] | None:
            if current_uid in path:
                cycle_start = path.index(current_uid)
                return path[cycle_start:] + [current_uid]

            if current_uid in visited_inheritance:
                return None

            visited_inheritance.add(current_uid)
            path.append(current_uid)

            for base_uid in inheritance_graph.get(current_uid, set()):
                if base_uid not in visited_inheritance:
                    cycle = dfs(base_uid)
                    if cycle:
                        return cycle

            path.pop()
            return None

        return dfs(start_uid)

    for uid in classlike_uids:
        if uid not in visited_inheritance:
            cycle = find_inheritance_cycle(uid)
            if cycle:
                cycle_str = " -> ".join(f"{get_name(uid_to_def[u])} ({u.resource})" for u in cycle)
                raise ValueError(f"Inheritance cycle detected: {cycle_str}")

    if LOG_TS:
        hprint(DEP_GRAPH, "No inheritance cycles detected")

    # ===== STAGE 2: Create strata from dependency cycles =====

    def find_dependency_cycles() -> list[tuple[set[DefnUID], set[DefnUID]]]:
        """Find all strongly connected components with multiple classlike nodes"""
        index_counter = 0
        stack: list[DefnUID] = []
        lowlinks: dict[DefnUID, int] = {}
        index: dict[DefnUID, int] = {}
        on_stack: set[DefnUID] = set()
        sccs: list[tuple[set[DefnUID], set[DefnUID]]] = []

        def strongconnect(uid: DefnUID) -> None:
            nonlocal index_counter
            index[uid] = index_counter
            lowlinks[uid] = index_counter
            index_counter += 1
            stack.append(uid)
            on_stack.add(uid)

            for dep_uid in dependency_graph.get(uid, set()):
                if dep_uid not in index:
                    strongconnect(dep_uid)
                    lowlinks[uid] = min(lowlinks[uid], lowlinks[dep_uid])
                elif dep_uid in on_stack:
                    lowlinks[uid] = min(lowlinks[uid], index[dep_uid])

            if lowlinks[uid] == index[uid]:
                scc: set[DefnUID] = set()
                while True:
                    w = stack.pop()
                    on_stack.remove(w)
                    scc.add(w)
                    if w == uid:
                        break
                classlike_in_scc = scc.intersection(classlike_uids)
                if len(classlike_in_scc) >= 1:
                    sccs.append((classlike_in_scc, scc))
                elif len(scc) > 1:
                    raise RuntimeError(
                        f"Found a dependency component with multiple nodes but none are classlike: {scc}"
                    )
                else:
                    # Singleton SCC
                    sccs.append((classlike_in_scc, scc))

        for uid in dependency_graph.keys():
            if uid not in index:
                strongconnect(uid)

        return sccs

    all_sccs = find_dependency_cycles()

    uid_to_scc: dict[DefnUID, frozenset[DefnUID]] = {}
    scc_to_stratum: dict[frozenset[DefnUID], frozenset[DefnUID]] = {}
    classlike_scc_count = 0

    for classlike_scc, full_scc in all_sccs:
        frozen_scc = frozenset(full_scc)
        for uid in full_scc:
            uid_to_scc[uid] = frozen_scc

        if len(classlike_scc) > 0:
            classlike_scc_count += 1
            stratum = frozenset(classlike_scc)
            scc_to_stratum[frozen_scc] = stratum

    # Build SCC dependency graph (for all SCCs)
    all_scc_set = set(uid_to_scc.values())
    scc_deps: dict[frozenset[DefnUID], set[frozenset[DefnUID]]] = {}
    for scc in all_scc_set:
        deps: set[frozenset[DefnUID]] = set()
        for uid in scc:
            for dep_uid in dependency_graph.get(uid, set()):
                dep_scc = uid_to_scc[dep_uid]
                if dep_scc != scc:
                    deps.add(dep_scc)
        scc_deps[scc] = deps

    if LOG_TS:
        hprint(
            DEP_GRAPH, f"Found {len(all_scc_set)} SCCs ({classlike_scc_count} with classlike nodes)"
        )

    # ===== STAGE 3: Sort strata by dependency order =====

    # Toposort SCCs and collect strata
    sorted_strata: list[frozenset[DefnUID]] = []
    visited_sccs: set[frozenset[DefnUID]] = set()

    def toposort_scc_dfs(scc: frozenset[DefnUID]) -> None:
        if scc in visited_sccs:
            return
        visited_sccs.add(scc)
        for dep_scc in scc_deps[scc]:
            toposort_scc_dfs(dep_scc)
        if scc in scc_to_stratum:
            sorted_strata.append(scc_to_stratum[scc])

    for scc in all_scc_set:
        toposort_scc_dfs(scc)

    if LOG_TS:
        hprint(DEP_GRAPH, f"Sorted {len(sorted_strata)} strata")

    # ===== STAGE 4: Process each stratum in order =====

    ordered_defs: list[DefUniMsg] = []
    visited_uids: set[DefnUID] = set()

    def sort_by_inheritance(stratum_uids: set[DefnUID]) -> list[DefnUID]:
        """Sort class-like defs within a stratum by inheritance order."""
        stratum_ordered: list[DefnUID] = []
        visited_in_stratum: set[DefnUID] = set()

        def inheritance_dfs(uid: DefnUID) -> None:
            if uid in visited_in_stratum:
                return
            visited_in_stratum.add(uid)
            bases = inheritance_graph.get(uid, set())
            bases_in_stratum = bases.intersection(stratum_uids)
            for base_uid in bases_in_stratum:
                inheritance_dfs(base_uid)
            stratum_ordered.append(uid)

        for uid in stratum_uids:
            inheritance_dfs(uid)

        return stratum_ordered

    for stratum in sorted_strata:
        stratum_ordered_uids = sort_by_inheritance(set(stratum))

        for classlike_uid in stratum_ordered_uids:
            collected_non_classlike: list[DefnUID] = []
            forward_classlike_refs: dict[DefnUID, str] = {}

            def collect_and_find_forward_refs(current_uid: DefnUID) -> None:
                """DFS collecting non-classlike nodes and recording forward classlike refs."""
                if current_uid in visited_uids:
                    return

                if current_uid in classlike_uids:
                    forward_classlike_refs[current_uid] = get_name(uid_to_def[current_uid])
                    return

                for dep_uid in dependency_graph.get(current_uid, set()):
                    collect_and_find_forward_refs(dep_uid)

                collected_non_classlike.append(current_uid)
                visited_uids.add(current_uid)

            # Note: dependency graph contains self references if there are any
            for dep_uid in dependency_graph.get(classlike_uid, set()):
                collect_and_find_forward_refs(dep_uid)

            if forward_classlike_refs and LOG_TS:
                hprint(
                    DEP_GRAPH,
                    f"Replacing forward refs in {get_name(uid_to_def[classlike_uid])}:\n"
                    f"{list(forward_classlike_refs.keys())} -> {list(forward_classlike_refs.values())}",
                )

            # Replace forward RefMsg in the collected defs
            for uid in collected_non_classlike:
                uid_to_def[uid] = replace_refs_in_def(uid_to_def[uid], forward_classlike_refs)

            # Replace forward RefMsg in the classlike def itself
            updated_classlike = replace_refs_in_def(
                uid_to_def[classlike_uid], forward_classlike_refs
            )
            uid_to_def[classlike_uid] = updated_classlike

            ordered_defs.extend(uid_to_def[uid] for uid in collected_non_classlike)
            ordered_defs.append(updated_classlike)
            visited_uids.add(classlike_uid)

    # ===== STAGE 5: Append remaining non-classlike defs =====

    for uid in uid_to_def.keys():
        if uid not in visited_uids:
            ordered_defs.append(uid_to_def[uid])
            visited_uids.add(uid)

    ordered_keys = []
    if keys:
        assert len(defs) == len(keys), "defs and keys must have the same length"
        key_map = {definition.uid: key for key, definition in zip(keys, defs)}
        ordered_keys = [key_map[definition.uid] for definition in ordered_defs]

    return ordered_defs, ordered_keys
